from torch import distributed as dist
import torch
import os
import signal
import asyncio
from functools import wraps
from torch import multiprocessing as mp
from torch.nn.parallel import DistributedDataParallel as DDP
from contextlib import contextmanager
from torch.nn.parallel.distributed import _find_tensors
from .contextlib import maybeasynccontextmanager
import logging
import inspect
import time

log = logging.getLogger(__name__)

def initialize(device, devices):
    if dist.is_initialized():
        log.info('Process group is already initialized')
        return 

    os.environ['MASTER_ADDR'] = '127.0.0.1'
    if device is None:
        os.environ['MASTER_PORT'] = str(29500)
        dist.init_process_group('nccl', rank=0, world_size=1)
    else:
        os.environ['MASTER_PORT'] = str(29500 + devices[0])
        dist.init_process_group('nccl', rank=devices.index(device), world_size=len(devices))

@contextmanager
def processgroup(device, devices):
    try:
        initialize(device, devices)
        yield
    finally:
        dist.destroy_process_group()

class DDP2(DDP):

    def forward(self, *inputs, **kwargs):
        """Modified to support not scattering inputs when it's only on one device"""
        if self.require_forward_param_sync:
            self._sync_params()

        if self.device_ids:
            if len(self.device_ids) == 1:
                output = self.module(*inputs, **kwargs)
            else:
                inputs, kwargs = self.scatter(inputs, kwargs, self.device_ids)
                outputs = self.parallel_apply(self._module_copies[:len(inputs)], inputs, kwargs)
                output = self.gather(outputs, self.output_device)
        else:
            output = self.module(*inputs, **kwargs)

        if torch.is_grad_enabled() and self.require_backward_grad_sync:
            self.require_forward_param_sync = True
            # We'll return the output object verbatim since it is a freeform
            # object. We need to find any tensors in this object, though,
            # because we need to figure out which parameters were used during
            # this forward pass, to ensure we short circuit reduction for any
            # unused parameters. Only if `find_unused_parameters` is set.
            if self.find_unused_parameters:
                self.reducer.prepare_for_backward(list(_find_tensors(output)))
            else:
                self.reducer.prepare_for_backward([])
        else:
            self.require_forward_param_sync = False

        return output

def set_start_method():
    """If you use get_start_method to check what the start method is, you'll accidentally _set_ the start method
    and then get an error if you later try to set it. Here we open the box without killing the cat"""
    import os

    # https://github.com/pytorch/pytorch/issues/32575
    # os.environ['NCCL_BLOCKING_WAIT'] = '1'

    from multiprocessing import context
    ctx = context._default_context
    if ctx._actual_context is None:
        mp.set_start_method('spawn')
    else:
        assert ctx._actual_context._name in ('spawn', 'forkserver')

def consensus(b):
    b = torch.tensor(float(b)).cuda()
    dist.all_reduce(b, dist.ReduceOp.PRODUCT)
    return bool(b.cpu())

def cancel(canceller):
    if dist.is_initialized():
        # If we're in a process group, either the whole group needs to 
        # break or no-one does, else a process will be left hanging.
        cancel = canceller.is_set()
        if cancel:
            log.info('Canceller set, trying to break')
        if consensus(cancel):
            log.info('Everyone has cancelled, breaking')
            return True
    else:
        if canceller.is_set():
            log.info('Cancelled, breaking')
            return True

async def surrender():
    await asyncio.sleep(0)


class DeadStrand(Exception):
    pass

def coroutine_runner(f, *args, **kwargs):
    co = f(*args, **kwargs)
    try:
        while True:
            co.send(None)
    except StopIteration:
        pass
    except Exception as e:
        raise e


class ProcessSentinel:

    def __init__(self, wait=15):
        self._wait = wait
        self._processes = {}
        self._references = []
        self.canceller = mp.Event()
        set_start_method()

        self.serial = False

    def pin(self, obj):
        """There are sometimes objects passed to children - like queues - that need to 
        stick around for as long as the children do"""
        self._references.append(obj)

    def launch(self, f, *args, **kwargs):
        if (self.canceller not in args) and (self.canceller not in kwargs.values()):
            log.warn('Sentinel\'s canceller has not been passed to a launched process')

        count = len([n for n in self._processes if n == f.__qualname__])
        if inspect.iscoroutinefunction(f):
            proc = mp.Process(
                name=f'{f.__qualname__}-{count}',
                target=coroutine_runner,
                args=(f, *args),
                kwargs=kwargs)
        else:
            proc = mp.Process(
                name=f'{f.__qualname__}-{count}',
                target=f,
                args=args,
                kwargs=kwargs)
        proc.start()
        self._processes[f.__qualname__, count] = proc
        log.info(f'Launched process {f.__qualname__}-{count}')

    def wait(self):
        for _ in range(int(self._wait)):
            alive = [(n, c) for (n, c), p in self._processes.items() if p.is_alive()]
            if alive:
                strs = [f'{n}-{c}' for n, c in alive]
                log.info(f'Waiting for cancellations: {", ".join(strs)} still alive')
            else:
                log.info('All processes gracefully cancelled')
                break
            time.sleep(1)
        else:
            for n, c in alive:
                log.info(f'Failed to cancel "{n}-{c}"; terminating')
                self._processes[n].terminate() 

        self._references = []

    def cancel(self):
        log.info('Setting canceller')
        self.canceller.set()
        self.wait()

    def check(self):
        for (n, c), p in self._processes.items():
            if not p.is_alive():
                log.info(f'Process "{n}-{c}" died unexpectedly; cancelling')
                self.cancel()
                raise DeadStrand(f'Process "{n}-{c}" died unexpectedly')

class SerialSentinel:

    def __init__(self, wait=15):
        self._wait = wait
        self.canceller = mp.Event()

        self._coroutines = {}
        self._exited = []

        self.serial = True

    def launch(self, f, *args, **kwargs):
        if (self.canceller not in args) and (self.canceller not in kwargs.values()):
            log.warn('Sentinel\'s canceller has not been passed to a launched process')

        count = len([n for n, _ in self._coroutines if n == f.__qualname__])

        co = f(*args, **kwargs)
        self._coroutines[f.__qualname__, count] = co
        log.info(f'Launched coroutine {f.__qualname__}-{count}')

    def wait(self):
        for _ in range(int(self._wait)):
            alive = []
            for (n, c), co in self._coroutines.items():
                try:
                    co.send(None)
                except (RuntimeError, StopIteration):
                    pass
                else:
                    alive.append((n, c))
            if alive:
                strs = [f'{n}-{c}' for n, c in alive]
                log.info(f'Waiting for cancellations: {", ".join(strs)} still alive')
            else:
                log.info('All coroutines gracefully cancelled')
                break
        else:
            for n, c in alive:
                log.info(f'Failed to cancel "{n}-{c}"; closing')
                try:
                    self._coroutines[n, c].close()
                except RuntimeError:
                    pass

    def cancel(self):
        log.info('Setting canceller')
        self.canceller.set()
        self.wait()

    def check(self):
        for (n, c), co in self._coroutines.items():
            try:
                co.send(None)
            except StopIteration:
                pass
            except Exception as e:
                log.info(f'Coroutine "{n}-{c}" died unexpectedly; cancelling')
                self.cancel()
                raise e

@contextmanager
def sentinel(serial=False):
    sentinel = SerialSentinel() if serial else ProcessSentinel()
    try:
        yield sentinel
    except KeyboardInterrupt:
        log.info('Got a keyboard interrupt, cancelling processes')
        sentinel.cancel()
    except (DeadStrand,):
        raise
    except:
        sentinel.cancel()  
        raise
    else:
        sentinel.cancel()



from tqdm.auto import tqdm
from contextlib import contextmanager
import multiprocessing
import types
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor, Future, _base, as_completed
import logging

log = logging.getLogger(__name__)

class SerialExecutor(_base.Executor):
    """An executor that runs things on the main process/thread - meaning stack traces are interpretable
    and the debugger works!
    """
    
    def __init__(self, *args, **kwargs):
        pass
    
    def submit(self, f, *args, **kwargs):
        future = Future()
        future.set_result(f(*args, **kwargs))
        return future

@contextmanager
def VariableExecutor(N=None, processes=True, **kwargs):
    """An executor that can be easily switched between serial, thread and parallel execution.
    If N=0, a serial executor will be used.
    """
    
    N = multiprocessing.cpu_count() if N is None else N
    
    if N == 0:
        executor = SerialExecutor
    elif processes:
        executor = ProcessPoolExecutor
    else:
        executor = ThreadPoolExecutor
    
    log.debug('Launching a {} with {} processes'.format(executor.__name__, N))    
    with executor(N, **kwargs) as pool:
        yield pool
        
@contextmanager
def parallel(f, progress=True, **kwargs):
    """Sugar for using the VariableExecutor. Call as
    
    with parallel(f) as g:
        ys = g.wait({x: g(x) for x in xs})
    and f'll be called in parallel on each x, and the results collected in a dictionary.
    A fantastic additonal feature is that if you pass `parallel(f, N=0)` , everything will be run on 
    the host process, so you can `import pdb; pdb.pm()` any errors. 
    """

    with VariableExecutor(**kwargs) as pool:

        def reraise(f, futures={}):
            e = f.exception()
            if e:
                log.warn('Exception raised on "{}"'.format(futures[f]), exc_info=e)
                raise e
            return f.result()

        submitted = set()

        def submit(*args, **kwargs):
            fut = pool.submit(f, *args, **kwargs)
            submitted.add(fut)
            fut.add_done_callback(submitted.discard)  # Try to avoid memory leak
            return fut
        
        def wait(c):
            # Recurse on list-likes
            if type(c) in (list, tuple, types.GeneratorType):
                ctor = list if isinstance(c, types.GeneratorType) else type(c)
                results = wait(dict(enumerate(c)))
                return ctor(results[k] for k in sorted(results))

            # Now can be sure we've got a dict-like
            futures = {fut: k for k, fut in c.items()}
            
            results = {}
            for fut in tqdm(as_completed(futures), total=len(c), disable=not progress):
                results[futures[fut]] = reraise(fut, futures)
                
            return results
        
        def cancel():
            while True:
                remaining = list(submitted)
                for fut in remaining:
                    fut.cancel()
                    submitted.discard(fut)
                if not remaining:
                    break

        try:
            submit.wait = wait
            yield submit
        finally:
            cancel()


"""TODO-DOCS learning docs"""
import torch
import numpy as np

def batch_indices(chunk, batch_size):
    T, B = chunk.world.reset.shape
    batch_width = batch_size//T
    indices = torch.randperm(B, device=chunk.world.reset.device)
    indices = [indices[i:i+batch_width] for i in range(0, B, batch_width)]
    return indices

def gather(arr, indices):
    if isinstance(arr, dict):
        return arr.__class__({k: gather(arr[k], indices[k]) for k in arr})
    return torch.gather(arr, -1, indices.type(torch.long).unsqueeze(-1)).squeeze(-1)

def flatten(arr):
    if isinstance(arr, dict):
        return torch.cat([flatten(v) for v in arr.values()], -1)
    return arr

def assert_same_shape(ref, *arrs):
    for a in arrs:
        assert ref.shape == a.shape

def deltas(value, reward, target, reset, gamma=.99):
    reward, reset = reward[1:], reset[1:]
    regular_deltas = (reward + gamma*target[1:]) - value[:-1]
    return torch.where(reset, reward - value[:-1], regular_deltas)

def present_value(dv, finals, reset, alpha):
    assert_same_shape(dv, reset)

    reset = reset.type(torch.float)
    acc = finals
    result = torch.full_like(dv, np.nan)
    for t in np.arange(dv.shape[0])[::-1]:
        acc = dv[t] + acc*alpha*(1 - reset[t])
        result[t] = acc
    return result

def generalized_advantages(value, reward, v, reset, gamma, lambd=.97):
    assert_same_shape(value, reward, v, reset)

    dv = deltas(value, reward, v, reset, gamma=gamma)
    finals = torch.zeros_like(dv[-1])
    return torch.cat([present_value(dv, finals, reset[1:], lambd*gamma), finals[None]], 0).detach()

def reward_to_go(reward, value, reset, gamma):
    return torch.cat([present_value(reward[1:], value[-1], reset[1:], gamma), value[[-1]]], 0).detach()

def v_trace(ratios, value, reward, reset, gamma, max_rho=1, max_c=1):
    assert_same_shape(ratios, value, reward, reset)

    rho = ratios.clamp(0, max_rho)
    c = ratios.clamp(0, max_c)
    dV = rho[:-1]*deltas(value, reward, value, reset, gamma=gamma)

    discount = (1 - reset.int())[1:]*gamma

    A = value[:-1] + dV - discount*c[:-1]*value[1:]
    B = discount*c[:-1]

    v = torch.zeros_like(value)
    v[-1] = value[-1]
    for t in reversed(range(len(v)-1)):
        v[t] = A[t] + B[t]*v[t+1]

    return v.detach()

#########
# TESTS #
#########

def v_trace_ref(ratios, value, reward, reset, gamma=.99, max_rho=1, max_c=1):
    rho = ratios.clamp(0, max_rho)
    c = ratios.clamp(0, max_c)

    v = value.clone()
    for s in range(len(v)-1):
        for t in range(s, len(v)-1):
            prod_c = c[s:t].prod()
            if reset[t+1]:
                # If the next state is terminal, then the next value is zero
                dV = rho[t]*(reward[t+1] - value[t])
                v[s] += gamma**(t - s) * prod_c*dV
                break
            else:
                dV = rho[t]*(reward[t+1] + gamma*value[t+1] - value[t])
                v[s] += gamma**(t - s) * prod_c*dV
    return v

def test_v_trace():
    ratios = torch.tensor([1., 1., 1.])
    reward = torch.tensor([1., 2., 3.])
    value = torch.tensor([4., 5., 6.])
    gamma = 1.

    reset = torch.tensor([False, False, False])
    actual = v_trace(ratios, value, reward, reset, gamma)
    torch.testing.assert_allclose(actual, torch.tensor([11., 9., 6.]))

    reset = torch.tensor([False, True, False])
    actual = v_trace(ratios, value, reward, reset, gamma)
    torch.testing.assert_allclose(actual, torch.tensor([2., 9., 6.]))

def test_v_trace_ref():
    ratios = torch.tensor([1., 1., 1.])
    reward = torch.tensor([1., 2., 3.])
    value = torch.tensor([4., 5., 6.])
    gamma = 1.

    reset = torch.tensor([False, False, False])
    actual = v_trace_ref(ratios, value, reward, reset, gamma)
    torch.testing.assert_allclose(actual, torch.tensor([11., 9., 6.]))

    reset = torch.tensor([False, True, False])
    actual = v_trace_ref(ratios, value, reward, reset, gamma)
    torch.testing.assert_allclose(actual, torch.tensor([2., 9., 6.]))

def test_v_trace_equivalent(R=100, T=10):
    for _ in range(R):
        ratios = torch.rand((T,))
        value = torch.rand((T,))
        reward = torch.rand((T,))
        reset = torch.rand((T,)) > .8
        gamma = torch.rand(())

        expected = v_trace_ref(ratios, value, reward, reset, gamma)
        actual = v_trace(ratios, value, reward, reset, gamma)

        torch.testing.assert_allclose(expected, actual)

def test_reward_to_go():
    reward = torch.tensor([1., 2., 3.])
    value = torch.tensor([4., 5., 6.])
    gamma = 1.

    reset = torch.tensor([False, False, False])
    actual = reward_to_go(reward, value, reset, gamma)
    torch.testing.assert_allclose(actual, torch.tensor([11., 9., 6.]))

    reset = torch.tensor([False, True, False])
    actual = reward_to_go(reward, value, reset, gamma)
    torch.testing.assert_allclose(actual, torch.tensor([2., 9., 6.]))

def test_generalized_advantages():
    reward = torch.tensor([1., 2., 3])
    value = torch.tensor([4., 5., 6.])
    gamma = 1.
    lambd = 1.

    reset = torch.tensor([False, False, False])
    adv = generalized_advantages(value, reward, value, reset, gamma=gamma, lambd=lambd)
    torch.testing.assert_allclose(adv, torch.tensor([7., 4., 0.]))

    reset = torch.tensor([False, True, False])
    adv = generalized_advantages(value, reward, value, reset, gamma=gamma, lambd=lambd)
    torch.testing.assert_allclose(adv, torch.tensor([-2., 4., 0.]))


"""
TODO-DOCS Heads docs
"""
import numpy as np
import torch
from torch import nn
from torch.nn import functional as F

class MultiVectorIntake(nn.Module):

    def __init__(self, space, width):
        super().__init__()
        A, C = space.shape

        self.core = nn.Sequential(
                        nn.Linear(C, width), nn.ReLU(),)
        self.proj = nn.Sequential(
                        nn.Linear(A*width, width), nn.ReLU(),)
        
    def forward(self, obs, **kwargs):
        if obs.ndim == 3:
            return self.forward(obs[None], **kwargs).squeeze(0)

        T, B, A, C = obs.shape
        x = self.core(obs.reshape(T*B*A, C)).reshape(T, B, -1)
        return self.proj(x)

class MultiImageIntake(nn.Module):

    def __init__(self, space, width):
        super().__init__()
        A, C, H, W = space.shape

        self.conv = nn.Sequential(
                        nn.Conv2d(C, 32, (1, 8), stride=(1, 4)), nn.ReLU(),
                        nn.Conv2d(32, 64, (1, 4), stride=(1, 2)), nn.ReLU(),
                        nn.Conv2d(64, 128, (1, 3), stride=(1, 2)), nn.ReLU())

        zeros = torch.zeros((A, C, H, W))
        convwidth = self.conv(zeros).nelement()

        self.proj = nn.Sequential(
                        nn.Linear(convwidth, width), nn.ReLU(),
                        nn.Linear(width, width), nn.ReLU())

    def forward(self, obs, **kwargs):
        if obs.ndim == 5:
            return self.forward(obs[None], **kwargs).squeeze(0)

        T, B, A, C, H, W = obs.shape
        if obs.dtype == torch.uint8:
            obs = obs/255.
        x = self.conv(obs.reshape(T*B*A, C, H, W)).reshape(T, B, -1)
        return self.proj(x)

class ConcatIntake(nn.Module):

    def __init__(self, space, width):
        super().__init__()

        intakes = type(space)({k: intake(v, width) for k, v in space.items()})
        self.core = nn.Linear(len(intakes)*width, width)
        self.intakes = nn.ModuleDict(intakes)

    def forward(self, x, **kwargs):
        ys = [self.intakes[k](x[k]) for k in self.intakes]
        return self.core(torch.cat(ys, -1))

def intake(space, width):
    if isinstance(space, dict):
        return ConcatIntake(space, width)
    name = f'{type(space).__name__}Intake'
    if name in globals():
        return globals()[name](space, width)
    raise ValueError(f'Can\'t handle {space}')

class MultiDiscreteOutput(nn.Module):

    def __init__(self, space, width):
        super().__init__()
        shape = space.shape
        self.core = nn.Linear(width, int(np.prod(shape)))
        self.shape = shape
    
    def forward(self, x, **kwargs):
        y = self.core(x).reshape(*x.shape[:-1], *self.shape)
        return F.log_softmax(y, -1)

    def sample(self, logits, test=False):
        if test:
            return logits.argmax(-1)
        else:
            return torch.distributions.Categorical(logits=logits).sample()

class DictOutput(nn.Module):

    def __init__(self, space, width):
        super().__init__()
        self.core = nn.Linear(width, width*len(space))

        self._dtype = type(space)
        self.outputs = nn.ModuleDict({k: output(v, width) for k, v in space.items()})

    def forward(self, x, **kwargs):
        ys = torch.chunk(self.core(x), len(self.outputs), -1)
        return self._dtype({k: v(ys[i]) for i, (k, v) in enumerate(self.outputs.items())})
    
    def sample(self, l):
        return self._dtype({k: v.sample(l[k]) for k, v in self.outputs.items()})

class ValueOutput(nn.Module):

    def __init__(self, width):
        super().__init__()
        self.core = nn.Linear(width, 1)

    def forward(self, x, **kwargs):
        return self.core.forward(x).squeeze(-1)

def output(space, width):
    if isinstance(space, dict):
        return DictOutput(space, width)
    name = f'{type(space).__name__}Output'
    if name in globals():
        return globals()[name](space, width)
    raise ValueError(f'Can\'t handle {space}')




from io import BytesIO
import logging
import requests
from tqdm.auto import tqdm
from zipfile import ZipFile
import pandas as pd
from pathlib import Path
import gzip
import numpy as np
from rebar import parallel, dotdict
import ast
from pkg_resources import resource_filename
from pathlib import Path

log = logging.getLogger(__name__)

LICENSE = """The cubicasa geometry is based on the Cubicasa5k dataset. 
While megastep as a whole can be used for any purpose, this geometry 
specifically is offered for non-commercial use only. You can read 
more about this - and about alternatives - in the megastep FAQ at 

http://andyljones.com/megastep/faq
"""

REJECTION = """
You entered "{i}". You cannot download the cubicasa dataset 
without confirming you understand the licensing conditions. Please read 
the FAQ for suggestions about alternative geometries. 

If your problem is specifically entering the 'Y' character, you can run 
this snippet from a console:
```
python -c "from megastep import cubicasa; cubicasa.force_confirm()
```
"""

PATH = Path(resource_filename(__package__, '.cubicasa-confirmed'))

def confirm():
    if PATH.exists():
        return 

    print(LICENSE)
    print('Please enter "Y" to confirm you understand this.\n', flush=True)
    i = input('[Y/N]: ')
    if i in 'yY':
        PATH.touch()
        print('\nConfirmed.')
        return 
    else:
        print(REJECTION.format(i=i))
        raise ValueError('You refused to confirm that you understand the license restrictions.')

def force_confirm():
    """:ref:`As described in the FAQ <cubicasa-license>, the cubicasa dataset has a non-commercial use restriction`. 
    
    Most users will be prompted for their understanding of this restriction when they first download the dataset, but
    this might cause trouble for people who are using automated systems, so calling this function 
    is offered as an alternative.
    """
    print(LICENSE)
    print('By calling `force_confirm`, you confirm you understand this.')
    PATH.touch()

def download(url):
    bs = BytesIO()
    log.info(f'Downloading {url}')
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        total = int(r.headers['Content-Length']) if 'Content-Length' in r.headers else None
        with tqdm(total=total, unit_scale=True, unit_divisor=1024, unit='B') as pbar:
            for chunk in r.iter_content(chunk_size=2**20): 
                pbar.update(len(chunk))
                bs.write(chunk)
    return bs.getvalue()

def cubicasa5k():
    p = Path('.cache/cubicasa.zip')
    if not p.exists():
        url = 'https://zenodo.org/record/2613548/files/cubicasa5k.zip?download=1'
        p.parent.mkdir(exist_ok=True, parents=True)
        p.write_bytes(download(url))
    return str(p)

def svg_data(regenerate=False):
    p = Path('.cache/cubicasa-svgs.json.gz')
    if not p.exists() or regenerate:
        p.parent.mkdir(exist_ok=True, parents=True)
        if regenerate:
            log.info('Regenerating SVG cache from cubicasa dataset. This will require a 5G download.')
            with ZipFile(cubicasa5k()) as zf:
                pattern = r'cubicasa5k/(?P<category>[^/]*)/(?P<id>\d+)/(?P<filename>[^.]*)\.svg'
                svgs = (pd.Series(zf.namelist(), name='path')
                            .to_frame()
                            .loc[lambda df: df.path.str.match(pattern)]
                            .reset_index(drop=True))
                svgs = pd.concat([svgs, svgs.path.str.extract(pattern)], axis=1)
                svgs['svg'] = svgs.path.apply(lambda p: zf.read(p).decode())
                compressed = gzip.compress(svgs.to_json().encode())
                p.write_bytes(compressed)
        else:
            #TODO: Shift this to Github 
            url = 'https://www.dropbox.com/s/iblduqobhqomz4g/cubicasa-svgs.json.gzip?raw=1'
            p.write_bytes(download(url))
    return pd.read_json(gzip.decompress(p.read_bytes()))

def flatten(tree):
    flat = {}
    for k, v in tree.items():
        if isinstance(v, dict):
            for kk, vv in flatten(v).items():
                flat[f'{k}/{kk}'] = vv
        else:
            flat[k] = v
    return flat

def unflatten(d):
    tree = type(d)()
    for k, v in d.items():
        parts = k.split('/')
        node = tree
        for p in parts[:-1]:
            node = node.setdefault(p, type(d)())
        node[parts[-1]] = v
    return tree
        
def safe_geometry(id, svg):
    try: 
        # Hide the import since it uses a fair number of libraries not used elsewhere.
        from . import geometry
        return geometry.geometry(svg)
    except:
        # We'll lose ~8 SVGs to them not having any spaces
        log.info(f'Geometry generation failed on on #{id}')

def fastload(raw):
    """Most of the time in np.load is spent parsing the header, since it could have a giant mess of record types in
    it. But we know here that it doesn't! So we can do the parsing with direct slices into the byte array, and skip a
    bunch of checks.
    
    Can push x3 faster than this by writing a regex for the descr and shape, but that's optimizing a bit too hard for 
    our purposes.
    
    Credit to @pag for pointing this out to me once upon a time"""
    headerlen = np.frombuffer(raw[8:9], dtype=np.uint8)[0]
    header = ast.literal_eval(raw[10:10+headerlen].decode())
    return np.frombuffer(raw[10+headerlen:], dtype=header['descr']).reshape(header['shape'])

def geometry_data(regenerate=False):
    # Why .npz.gz? Because applying gzip manually manages x10 better compression than
    # np.savez_compressed. They use the same compression alg, so I assume the difference
    # is in the default compression setting - which isn't accessible in np.savez_compressec.
    p = Path('.cache/cubicasa-geometry.npz.gz')
    if not p.exists() or regenerate:
        p.parent.mkdir(exist_ok=True, parents=True)
        if regenerate:
            log.info('Regenerating geometry cache from SVG cache.')
            with parallel.parallel(safe_geometry) as pool:
                gs = pool.wait({str(row.id): pool(row.id, row.svg) for _, row in svg_data().iterrows()})
            gs = flatten({k: v for k, v in gs.items() if v is not None})

            bs = BytesIO()
            np.savez(bs, **gs)
            p.write_bytes(gzip.compress(bs.getvalue()))
        else:
            #TODO: Shift this to Github 
            url = 'https://www.dropbox.com/s/3ohut8lvmr8lkwg/cubicasa-geometry.npz.gz?raw=1'
            p.write_bytes(download(url))

    # np.load is kinda slow. 
    raw = gzip.decompress(p.read_bytes())
    with ZipFile(BytesIO(raw)) as zf:
        flat = dotdict.dotdict({n[:-4]: fastload(zf.read(n)) for n in zf.namelist()})
    return unflatten(flat)

_cache = None
def sample(n_geometries, split='training', seed=1):
    """Returns a random sample of cubicasa :ref:`geometries <geometry>`. 

    If you pass the same arguments, you'll get the same sample every time.

    There are 4,992 unique geometries, split into a 4,492-geometry training set and a 500-geometry test set.

    **Caching**

    The geometries are derived from the `Cubicasa5k <https://github.com/CubiCasa/CubiCasa5k>`_ dataset.

    The first time you call this function, it'll fetch and cache a ~10MB precomputed geometries file. This is far
    easier to work with than the full 5GB Cuibcasa5k dataset. If you want to recompute the geometries from scratch
    however, import this module and try calling :: 
    
        svg_data(regenerate=True) 
        geometry_data(regenerate=True)

    **Parameters**

    :param n_designs: The number of geometries to return
    :type n_designs: int
    :param split: Whether to return a sample from the ``training`` set, the ``test`` set, or ``all`` . The split is
        90/10 in favour of the training set. Defaults to ``training`` . 
    :type split: str
    :param seed: The seed to use when allocating the training and test sets.

    :return: A list of geometries.
    """
    confirm()
    global _cache
    if _cache is None:
        _cache = geometry_data()
        # Add the ID, since we're going to return this as a list
        _cache = type(_cache)({k: type(v)({'id': k, **v}) for k, v in _cache.items()})
    
    cutoff = int(.9*len(_cache))
    order = np.random.RandomState(seed).permutation(sorted(_cache))
    if split == 'training':
        order = order[:cutoff]
    elif split == 'test':
        order = order[cutoff:]
    elif split == 'all':
        order = order
    else:
        raise ValueError('Split must be train/test/all')

    return [_cache[order[i % len(order)]] for i in range(n_geometries)]



import copy
import logging
import time
from pathlib import Path
from collections import defaultdict, deque
import logging.handlers
import ipywidgets as widgets
from contextlib import contextmanager
import psutil
from . import widgets, paths
from .contextlib import maybeasynccontextmanager
import sys
import traceback
import _thread
import threading

# for re-export
from logging import getLogger

log = getLogger(__name__)

#TODO: This shouldn't be at the top level
logging.basicConfig(
            stream=sys.stdout, 
            level=logging.INFO, 
            format='%(asctime)s %(levelname)s %(name)s: %(message)s', 
            datefmt=r'%Y-%m-%d %H:%M:%S')
logging.getLogger('parso').setLevel('WARN')  # Jupyter's autocomplete spams the output if this isn't set
log.info('Set log params')

def in_ipython():
    try:
        __IPYTHON__
        return True
    except NameError:
        return False

class StdoutRenderer:

    def __init__(self):
        super().__init__()

    def emit(self, path, line):
        source = '{procname}/#{pid}'.format(**paths.parse(path))
        print(f'{source}: {line}')

    def close(self):
        pass

class IPythonRenderer:

    def __init__(self, compositor=None):
        super().__init__()
        self._out = (compositor or widgets.Compositor()).output()
        self._next = time.time()
        self._lasts = {}
        self._buffers = defaultdict(lambda: deque(['']*self._out.lines, maxlen=self._out.lines))

    def _format_block(self, name):
        n_lines = max(self._out.lines//(len(self._buffers) + 2), 1)
        lines = '\n'.join(list(self._buffers[name])[-n_lines:])
        return f'{name}:\n{lines}'

    def _display(self, force=False):
        content = '\n\n'.join([self._format_block(n) for n in self._buffers])
        self._out.refresh(content)

        for name, last in list(self._lasts.items()):
            if time.time() - last > 120:
                del self._buffers[name]
                del self._lasts[name]

    def emit(self, path, line):
        source = '{procname}/#{pid}'.format(**paths.parse(path))
        self._buffers[source].append(line)
        self._lasts[source] = time.time()
        self._display()
    
    def close(self):
        self._display(force=True)
        # Want to leave the outputs open so you can see the final messages
        # self._out.close()
        super().close()

@contextmanager
def handlers(*new_handlers):
    logger = logging.getLogger()
    old_handlers = [*logger.handlers]
    try:
        logger.handlers = new_handlers
        yield 
    finally:
        for h in new_handlers:
            try:
                h.acquire()
                h.flush()
                h.close()
            except (OSError, ValueError):
                pass
            finally:
                h.release()

        logger.handlers = old_handlers

@maybeasynccontextmanager
def to_dir(run_name):
    path = paths.path(run_name, 'logs').with_suffix('.txt')
    handler = logging.FileHandler(path)
    handler.setLevel(logging.INFO)
    handler.setFormatter(logging.Formatter(
        fmt='%(asctime)s %(levelname)s %(name)s: %(message)s', 
        datefmt=r'%H:%M:%S'))

    with handlers(handler):
        try:
            yield
        except:
            log.info(f'Trace:\n{traceback.format_exc()}')
            raise

class Reader:

    def __init__(self, run_name):
        self._dir = paths.subdirectory(run_name, 'logs')
        self._files = {}

    def read(self):
        for path in self._dir.glob('*.txt'):
            if path not in self._files:
                self._files[path] = path.open('r')
        
        for path, f in self._files.items():
            for line in f.readlines():
                yield path, line.rstrip('\n')

def __from_dir(canceller, renderer, reader):
    while True:
        for path, line in reader.read():
            renderer.emit(path, line)

        if canceller.is_set():
            break

        time.sleep(.01)

def _from_dir(canceller, renderer, reader):
    try:
        __from_dir(canceller, renderer, reader)
    except KeyboardInterrupt:
        log.info('Interrupting main')
        _thread.interrupt_main()
        __from_dir(canceller, renderer, reader)

@contextmanager
def from_dir(run_name, compositor=None):
    if in_ipython():
        renderer = IPythonRenderer(compositor)
    else:
        renderer = StdoutRenderer()

    with to_dir(run_name):
        try:
            reader = Reader(run_name)
            canceller = threading.Event()
            thread = threading.Thread(target=_from_dir, args=(canceller, renderer, reader))
            thread.start()
            yield
        finally:
            log.info('Cancelling log forwarding thread')
            time.sleep(.25)
            canceller.set()
            thread.join(1)
            if thread.is_alive():
                log.error('Logging thread won\'t die')
            else:
                log.info('Log forwarding thread cancelled')

@contextmanager
def via_dir(run_name, compositor=None):
    with to_dir(run_name), from_dir(run_name, compositor):
        yield

### TESTS

def test_in_process():
    paths.clear('test', 'logs')

    with from_dir('test'):
        for _ in range(10):
            log.info('hello')
            time.sleep(.1)

def _test_multiprocess(run_name):
    with to_file(run_name):
        for i in range(10):
            log.info(str(i))
            time.sleep(.5)

def test_multiprocess():
    paths.clear('test', 'logs')

    import multiprocessing as mp
    with from_dir('test'):
        ps = []
        for _ in range(3):
            p = mp.Process(target=_test_multiprocess, args=('test',))
            p.start()
            ps.append(p)

        while any(p.is_alive() for p in ps):
            time.sleep(.5)

def _test_error(run_name):
    with to_file(run_name):
        log.info('Alive')
        time.sleep(2)
        raise ValueError('Last gasp')

def test_error():
    paths.clear('test', 'logs')

    import multiprocessing as mp
    with from_dir('test'):
        ps = []
        for _ in range(1):
            p = mp.Process(target=_test_error, args=('test',))
            p.start()
            ps.append(p)

        while any(p.is_alive() for p in ps):
            time.sleep(.5)



"""TODO-DOCS Demo docs"""
import torch
from megastep.demo import learning, lstm, heads
from rebar import logging, paths, stats, widgets, storing, arrdict, dotdict, recurrence, recording
import pandas as pd
import numpy as np
from torch import nn
from tqdm.auto import tqdm
from megastep.demo.envs import explorer, deathmatch

log = logging.getLogger(__name__)

class Agent(nn.Module):

    def __init__(self, env, width=256):
        super().__init__()
        out = heads.output(env.action_space, width)
        self.sampler = out.sample
        self.policy = recurrence.Sequential(
            heads.intake(env.obs_space, width),
            lstm.LSTM(d_model=width),
            out)
        self.value = recurrence.Sequential(
            heads.intake(env.obs_space, width),
            lstm.LSTM(d_model=width),
            heads.ValueOutput(width))

    def forward(self, world, sample=False, value=False, test=False):
        outputs = arrdict.arrdict(
            logits=self.policy(world.obs, reset=world.reset))
        if sample or test:
            outputs['actions'] = self.sampler(outputs.logits, test)
        if value:
            outputs['value'] = self.value(world.obs, reset=world.reset)
        return outputs

def as_chunk(buffer):
    chunk = arrdict.stack(buffer)
    with stats.defer():
        stats.rate('sample-rate/actor', chunk.world.reset.nelement())
        stats.mean('traj-length', chunk.world.reset.nelement(), chunk.world.reset.sum())
        stats.cumsum('count/traj', chunk.world.reset.sum())
        stats.cumsum('count/world', chunk.world.reset.size(0))
        stats.cumsum('count/chunks', 1)
        stats.cumsum('count/samples', chunk.world.reset.nelement())
        stats.rate('step-rate/chunks', 1)
        stats.rate('step-rate/world', chunk.world.reset.size(0))
        stats.mean('step-reward', chunk.world.reward.sum(), chunk.world.reward.nelement())
        stats.mean('traj-reward/mean', chunk.world.reward.sum(), chunk.world.reset.sum())
        stats.mean('traj-reward/positive', chunk.world.reward.clamp(0, None).sum(), chunk.world.reset.sum())
        stats.mean('traj-reward/negative', chunk.world.reward.clamp(None, 0).sum(), chunk.world.reset.sum())
    return chunk

def optimize(agent, opt, batch, entropy=1e-2, gamma=.99, clip=.2):
    w, d0 = batch.world, batch.decision
    d = agent(w, value=True)

    logits = learning.flatten(d.logits)
    old_logits = learning.flatten(learning.gather(d0.logits, d0.actions)).sum(-1)
    new_logits = learning.flatten(learning.gather(d.logits, d0.actions)).sum(-1)
    ratio = (new_logits - old_logits).exp().clamp(.05, 20)

    v_target = learning.v_trace(ratio, d.value, w.reward, w.reset, gamma=gamma)
    v_clipped = d0.value + torch.clamp(d.value - d0.value, -10, +10)
    v_loss = .5*torch.max((d.value - v_target)**2, (v_clipped - v_target)**2).mean()

    adv = learning.generalized_advantages(d.value, w.reward, d.value, w.reset, gamma=gamma)
    normed_adv = (adv - adv.mean())/(1e-3 + adv.std())
    free_adv = ratio*normed_adv
    clip_adv = torch.clamp(ratio, 1-clip, 1+clip)*normed_adv
    p_loss = -torch.min(free_adv, clip_adv).mean()

    h_loss = (logits.exp()*logits).sum(-1).mean()
    loss = v_loss + p_loss + entropy*h_loss
    
    opt.zero_grad()
    loss.backward()
    torch.nn.utils.clip_grad_norm_(agent.policy.parameters(), 100.)
    torch.nn.utils.clip_grad_norm_(agent.value.parameters(), 100.)

    opt.step()

    kl_div = -(new_logits - old_logits).mean().detach()
    with stats.defer():
        stats.mean('loss/value', v_loss)
        stats.mean('loss/policy', p_loss)
        stats.mean('loss/entropy', h_loss)
        stats.mean('resid-var/v', (v_target - d.value).pow(2).mean(), v_target.pow(2).mean())
        stats.mean('rel-entropy', -(logits.exp()*logits).sum(-1).mean()/np.log(logits.shape[-1]))
        stats.mean('kl-div', kl_div) 

        stats.mean('v-target/mean', v_target.mean())
        stats.mean('v-target/std', v_target.std())

        stats.mean('adv/z-mean', adv.mean())
        stats.mean('adv/z-std', adv.std())
        stats.max('adv/z-max', adv.abs().max())

        stats.rate('sample-rate/learner', w.reset.nelement())
        stats.rate('step-rate/learner', 1)
        stats.cumsum('count/learner-steps', 1)
        # stats.rel_gradient_norm('rel-norm-grad', agent)

        stats.mean('param/gamma', gamma)
        stats.mean('param/entropy', entropy)

    return kl_div

def train():
    buffer_size = 32
    n_envs = 8*1024
    batch_size = 16*1024

    env = explorer.Explorer(n_envs)
    agent = Agent(env).cuda()
    opt = torch.optim.Adam(agent.parameters(), lr=3e-4, amsgrad=True)

    run_name = f'{pd.Timestamp.now():%Y-%m-%d %H%M%S} {type(env).__name__}'
    paths.clear(run_name)
    compositor = widgets.Compositor()
    with logging.via_dir(run_name, compositor), stats.via_dir(run_name, compositor):
        
        world = env.reset()
        while True:
            buffer = []
            state = recurrence.get(agent)
            for _ in range(buffer_size):
                decision = agent(world[None], sample=True, value=True).squeeze(0).detach()
                buffer.append(arrdict.arrdict(
                    world=world,
                    decision=decision))
                buffer = buffer[-buffer_size:]
                world = env.step(decision)
                log.info('actor stepped')

            chunk = as_chunk(buffer)
            
            for idxs in learning.batch_indices(chunk, batch_size):
                with recurrence.temp_clear_set(agent, state[:, idxs]):
                    kl = optimize(agent, opt, chunk[:, idxs])

                log.info('learner stepped')
                if kl > .02:
                    log.info('kl div exceeded')
                    break
            storing.store_latest(run_name, {'agent': agent}, throttle=60)
            stats.gpu.memory(0)
            stats.gpu.vitals(0, throttle=15)

def demo(run=-1, length=None, test=True, N=None, env=None, agent=None, d=0):
    env = explorer.Explorer(d+1) if env is None else env
    world = env.reset()
    if agent is None:
        agent = Agent(env).cuda()
        agent.load_state_dict(storing.load(run)['agent'], strict=False)

    world = env.reset()
    steps = 0
    with recording.ParallelEncoder(env.plot_state, N=N) as encoder, \
            tqdm(total=length) as pbar:
        while True:
            decision = agent(world[None], sample=True, test=test, value=True).squeeze(0)
            world = env.step(decision)
            steps += 1
            pbar.update(1)
            if length is None and world.reset.any():
                break
            state = env.state(d)
            encoder(arrdict.numpyify(arrdict.arrdict(**state, decision=decision)))
            if (steps == length):
                break
    encoder.notebook()
    return encoder


"""TODO-DOCS Scene docs"""
import matplotlib as mpl
import numpy as np
import torch
from . import core, ragged, plotting
from rebar import arrdict
import matplotlib.pyplot as plt

# Ten bland colors from https://medialab.github.io/iwanthue/
COLORS = [
    "#c185ae",
    "#73a171",
    "#5666a4",
    "#9f7c4a",
    "#809cd5",
    "#566e40",
    "#8e537b",
    "#4f9fa4",
    "#b56d66",
    "#5a728c"]

def lengths(lines):
    return ((lines[..., 0, :] - lines[..., 1, :])**2).sum(-1)**.5

def agent_model():
    corners = [
            [-.5, -1.], [+.5, -1.], 
            [+1., -.5], [+1., +.5],
            [+.5, +1.], [-.5, +1.],
            [-1., +.5], [-1., -.5]]
    n = len(corners)
    walls = [[corners[i], corners[(i+1) % n]] for i in range(n)]
    return core.AGENT_WIDTH/2*np.array(walls)

def agent_colors():
    k, g, r = '.25', 'g', 'r'
    colors = (k, g, k, r, k, r, k, g)
    return np.stack([mpl.colors.to_rgb(s) for s in colors])

def resolutions(lines):
    return np.ceil(lengths(lines)/core.TEXTURE_RES).astype(int)

def wall_pattern(n, l=.5, random=np.random):
    p = core.TEXTURE_RES/l
    jumps = random.choice(np.array([0., 1.]), p=np.array([1-p, p]), size=n)
    jumps = jumps*random.normal(size=n)
    value = .5 + .5*(jumps.cumsum() % 1)
    return value

def init_textures(agentlines, agentcolors, walls, random=np.random):
    colormap = np.array([mpl.colors.to_rgb(c) for c in COLORS])
    wallcolors = colormap[np.arange(len(walls)) % len(colormap)]
    colors = np.concatenate([agentcolors, wallcolors])

    texwidths = resolutions(np.concatenate([agentlines, walls]))
    starts = texwidths.cumsum() - texwidths

    indices = np.full(texwidths.sum(), 0)
    indices[starts] = 1
    indices = np.cumsum(indices) - 1
    textures = core.gamma_decode(colors[indices])

    # Gives walls an even pattern that makes depth perception easy
    pattern = wall_pattern(textures.shape[0], random=random)
    pattern[:sum(texwidths[:len(agentlines)])] = 1.
    textures = textures*pattern[:, None]

    return textures, texwidths

def random_lights(lights, random=np.random):
    return np.concatenate([
        lights,
        random.uniform(.5, 2., (len(lights), 1))], -1)

@torch.no_grad()
def scenery(geometries, n_agents=1, device='cuda', random=np.random): 
    agentlines = np.tile(agent_model(), (n_agents, 1, 1))
    agentcolors = np.tile(agent_colors(), (n_agents, 1))

    data = []
    for g in geometries:
        lights = random_lights(g.lights)
        lines = np.concatenate([agentlines, g.walls])
        textures, texwidths = init_textures(agentlines, agentcolors, g.walls, random) 
        data.append(arrdict.arrdict(
            lights=arrdict.arrdict(vals=lights, widths=len(lights)),
            lines=arrdict.arrdict(vals=lines, widths=len(lines)),
            textures=arrdict.arrdict(vals=textures, widths=texwidths)))
    data = arrdict.torchify(arrdict.cat(data)).to(device)
    
    lights = ragged.Ragged(**data['lights'])
    scenery = core.cuda.Scenery(
        n_agents=n_agents,
        lights=lights,
        lines=ragged.Ragged(**data['lines']),
        textures=ragged.Ragged(**data['textures']),
        model=arrdict.torchify(agent_model()).to(device))
    core.cuda.bake(scenery)

    return scenery

def display(scenery, e=0):
    ax = plt.axes()

    state = arrdict.numpyify(
        arrdict.arrdict(
            scenery=scenery.state(e)))

    plotting.plot_lines(ax, state, zoom=False)
    plotting.plot_lights(ax, state)

    plotting.adjust_view(ax, state, zoom=False)

    return ax.figure


import signal
from .contextlib import maybeasynccontextmanager
import logging

log = logging.getLogger(__name__)

class Interrupter:

    def __init__(self):
        self._is_set = False

    def check(self):
        if self._is_set:
            self.reset()
            raise KeyboardInterrupt()

    def handle(self, signum, frame):
        log.info('Setting interrupt flag')
        self._is_set = True

    def reset(self):
        self._is_set = False

_INTERRUPTER = Interrupter()

@maybeasynccontextmanager
def interrupter():
    old = signal.signal(signal.SIGINT, _INTERRUPTER.handle)
    try:
        yield _INTERRUPTER
    finally:
        _INTERRUPTER.reset()
        signal.signal(signal.SIGINT, old)


import time
import numpy as np
import pandas as pd
from bokeh import plotting as bop, io as boi
from bokeh import models as bom, events as boe, layouts as bol
from bokeh.palettes import Category10_10
from itertools import cycle
from . import stats, paths
from contextlib import contextmanager
from IPython.display import clear_output

bop.output_notebook(hide_banner=True)

def array(fig):
    fig.canvas.draw_idle()
    renderer = fig.canvas.get_renderer()
    w, h = int(renderer.width), int(renderer.height)
    return (np.frombuffer(renderer.buffer_rgba(), np.uint8)
                        .reshape((h, w, 4))
                        [:, :, :3]
                        .copy())

def timedelta_xaxis(f):
    f.xaxis.ticker = bom.tickers.DatetimeTicker()
    f.xaxis.formatter = bom.FuncTickFormatter(code="""
        // TODO: Add support for millis

        // Calculate the hours, mins and seconds
        var s = Math.floor(tick / 1e3);
        
        var m = Math.floor(s/60);
        var s = s - 60*m;
        
        var h = Math.floor(m/60);
        var m = m - 60*h;
        
        var h = h.toString();
        var m = m.toString();
        var s = s.toString();
        var pm = m.padStart(2, "0");
        var ps = s.padStart(2, "0");

        // Figure out what the min resolution is going to be
        var min_diff = Infinity;
        for (var i = 0; i < ticks.length-1; i++) {
            min_diff = Math.min(min_diff, ticks[i+1]-ticks[i]);
        }

        if (min_diff <= 60e3) {
            var min_res = 2;
        } else if (min_diff <= 3600e3) {
            var min_res = 1;
        } else {
            var min_res = 0;
        }

        // Figure out what the max resolution is going to be
        if (ticks.length > 1) {
            var max_diff = ticks[ticks.length-1] - ticks[0];
        } else {
            var max_diff = Infinity;
        }

        if (max_diff >= 3600e3) {
            var max_res = 0;
        } else if (max_diff >= 60e3) {
            var max_res = 1;
        } else {
            var max_res = 2;
        }

        // Format the timedelta. Finally.
        if ((max_res == 0) && (min_res == 0)) {
            return `${h}h`;
        } else if ((max_res == 0) && (min_res == 1)) {
            return `${h}h${pm}`;
        } else if ((max_res == 0) && (min_res == 2)) {
            return `${h}h${pm}m${ps}`;
        } else if ((max_res == 1) && (min_res == 1)) {
            return `${m}m`;
        } else if ((max_res == 1) && (min_res == 2)) {
            return `${m}m${ps}`;
        } else if ((max_res == 2) && (min_res == 2)) {
            return `${s}s`;
        }
    """)

def suffix_yaxis(f):
    f.yaxis.formatter = bom.FuncTickFormatter(code="""
        var min_diff = Infinity;
        for (var i = 0; i < ticks.length-1; i++) {
            min_diff = Math.min(min_diff, ticks[i+1]-ticks[i]);
        }

        var suffixes = [
            'y', 'z', 'a', 'f', 'p', 'n', 'µ', 'm',
            '', 
            'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'];
        var precision = Math.floor(Math.log10(min_diff));
        var scale = Math.floor(precision/3);
        var index = scale + 8;
        if (index < 0) {
            //TODO: Fall back to numbro here
            return tick;
        } else if (index == 7) {
            // Millis are weird. Feels better to rende them as decimals.
            var decimals = -precision;
            return `${tick.toFixed(decimals)}`
        } else if (index < suffixes.length) {
            var suffix = suffixes[index];
            var scaled = tick/Math.pow(10, 3*scale);
            return `${scaled.toFixed(0)}${suffix}`
        } else {
            //TODO: Fall back to numbro here
            return tick;
        }
    """)

def x_zeroline(f):
    f.add_layout(bom.Span(location=0, dimension='height'))

def default_tools(f):
    f.toolbar_location = None
    f.toolbar.active_drag = f.select_one(bom.BoxZoomTool)
    # f.toolbar.active_scroll = f.select_one(bom.WheelZoomTool)
    # f.toolbar.active_inspect = f.select_one(bom.HoverTool)
    f.js_on_event(
        boe.DoubleTap, 
        bom.callbacks.CustomJS(args=dict(p=f), code='p.reset.emit()'))

def styling(f):
    timedelta_xaxis(f)
    suffix_yaxis(f)

def _timeseries(source, x, y):
    #TODO: Work out how to apply the axes formatters to the tooltips
    f = bop.figure(x_range=bom.DataRange1d(start=0, follow='end'), tooltips=[('', '$data_y')])
    f.line(x=x, y=y, source=source)
    default_tools(f)
    x_zeroline(f)
    styling(f)

    return f

def timeseries(s):
    source = bom.ColumnDataSource(s.reset_index())
    return _timeseries(source, s.index.name, s.name)

def _timedataframe(source, x, ys):
    f = bop.figure(x_range=bom.DataRange1d(start=0, follow='end'), tooltips=[('', '$data_y')])

    for y, color in zip(ys, cycle(Category10_10)):
        f.line(x=x, y=y, legend_label=y, color=color, width=2, source=source)

    default_tools(f)
    x_zeroline(f)
    styling(f)

    f.legend.label_text_font_size = '8pt'
    f.legend.margin = 7
    f.legend.padding = 0
    f.legend.spacing = 0
    f.legend.background_fill_alpha = 0.3
    f.legend.border_line_alpha = 0.
    f.legend.location = 'top_left'

    return f

def timedataframe(df):
    source = bom.ColumnDataSource(df.reset_index())
    return _timedataframe(source, df.index.name, df.columns)

def timegroups(df):
    tags = df.columns.str.extract(r'^(?P<chart1>.*)/(?P<label>.*)|(?P<chart2>.*)$')
    tags['chart'] = tags.chart1.combine_first(tags.chart2)
    tags.index = df.columns
    return tags[['chart', 'label']].fillna('')


class Stream:

    def __init__(self, run_name=-1, prefix=''):
        super().__init__()

        self._reader = stats.Reader(run_name, prefix)

        self._source = bom.ColumnDataSource({'time': np.array([0])})
        self._handle = None

    def _new_grid(self, children):
        return bol.gridplot(children, ncols=4, plot_width=350, plot_height=300, merge_tools=False)

    def _init(self, df):
        self._source = bom.ColumnDataSource(df.reset_index())

        children = []
        for name, group in timegroups(df).groupby('chart'):
            if group.label.eq('').all():
                assert len(group) == 1
                f = _timeseries(self._source, 'time', group.index[0])
                f.title = bom.Title(text=name)
            else:
                f = _timedataframe(self._source, 'time', group.index)
                f.title = bom.Title(text=name)
            children.append(f)
        self._grid = self._new_grid(children)
        ## TODO: Not wild about this
        clear_output(wait=True)
        self._handle = bop.show(self._grid, notebook_handle=True)

    def update(self, rule='60s', df=None):
        # Drop the last row as it'll be constantly refreshed as the period occurs
        df = self._reader.resample(rule).iloc[:-1] if df is None else df

        has_new_cols = not df.columns.isin(self._source.data).all()
        if has_new_cols:
            self._init(df)
        else:
            threshold = len(self._source.data['time'])
            new = df.iloc[threshold:]
            self._source.stream(new.reset_index())
        
        boi.push_notebook(handle=self._handle)

def view(run_name=-1, prefix='', rule='60s'):
    stream = Stream(run_name, prefix)
    while True:
        stream.update(rule=rule)
        time.sleep(1)

def review(run_name=-1, prefix='', rule='60s'):
    stream = Stream(run_name, prefix)
    stream.update(rule=rule)

def test_stream():
    times = pd.TimedeltaIndex([0, 60e3, 120e3])
    dfs = [
        pd.DataFrame([[0]], columns=['a'], index=times[:1]),
        pd.DataFrame([[0, 1], [10, 20]], columns=['a', 'b/a'], index=times[:2]),
        pd.DataFrame([[0, 1, 2], [10, 20, 30], [100, 200, 300]], columns=['a', 'b/a', 'b/b'], index=times[:3])]

    stream = Stream()
    for df in dfs:
        stream.update(df)
        time.sleep(1)
    


"""TODO-DOCS LSTM docs"""
import torch
from torch import nn
from rebar import recurrence, dotdict
from torch.nn.utils.rnn import PackedSequence

class Packer:

    def __init__(self, reset):
        self._reset = reset
        ext = reset.clone()
        ext[0] = True
        ext = ext.T.flatten()

        idxs = ext.nonzero().squeeze(-1)

        b = ext.cumsum(0)-1
        t = torch.arange(len(ext), device=idxs.device) - idxs[b]

        ends = torch.cat([idxs, torch.full_like(idxs[:1], len(b))])
        l = (ends[1:] - ends[:-1])

        T = reset.size(0)
        B = len(idxs)
        L = T+1
        assert T**2 * B < 2**32 - 1

        self._b = b
        self._t = t
        self._order = torch.argsort(t*B*L + b*L + (L-l[b]-1))
        self._sizes = torch.flip(torch.flip(torch.histc(l-1, l.max(), 0, l.max()), (0,)).cumsum(0), (0,))

    def pack_data(self, x):
        vals = x.transpose(0, 1).reshape(-1, *x.shape[2:])[self._order]
        return PackedSequence(vals, self._sizes.cpu())

    def pack_state(self, h):
        T, B = self._reset.shape
        J = self._sizes[0]
        mask = (self._order[:J] % T == 0) & ~self._reset.T.flatten()[self._order[:J]]
        hp = h.new_zeros((1, self._sizes[0], *h.shape[2:]))
        hp[:, mask] = h[:, ~self._reset[0]]
        return hp

    def pack(self, x, h, c):
        return self.pack_data(x), (self.pack_state(h), self.pack_state(c))

    def unpack_data(self, xp):
        T, B = self._reset.shape
        x = xp.data.new_zeros(T*B, *xp.data.shape[1:])
        x[self._order] = xp.data
        return x.reshape(B, T, *xp.data.shape[1:]).transpose(0, 1)

    def unpack_state(self, hp):
        T, B = self._reset.shape
        mask = (self._order % T == T-1)
        left_idxs = (self._order//T)[mask]
        right_idxs = self._b[self._order][mask]

        h = hp.new_zeros((1, B, *hp.shape[2:]))
        h[:, left_idxs] = hp[:, right_idxs]
        return h

    def unpack(self, xp, hcp):
        hp, cp = hcp
        return self.unpack_data(xp), (self.unpack_state(hp), self.unpack_state(cp))


class LSTM(nn.Module):

    def __init__(self, d_model):
        super().__init__()

        self._d_model = d_model
        self.lstm = nn.LSTM(input_size=d_model, hidden_size=d_model)
        self._h = recurrence.State()
        self._c = recurrence.State()

    def forward(self, x, reset):
        if x.ndim == 2:
            return self.forward(x[None], reset[None])[0]

        T, B, C = x.shape
        h0 = self._h.get(lambda: x.new_zeros(1, B, self._d_model))
        c0 = self._c.get(lambda: x.new_zeros(1, B, self._d_model))

        packer = Packer(reset)

        output = self.lstm(*packer.pack(x, h0, c0))
        y, (hn, cn) = packer.unpack(*output)
        self._h.set(hn.detach())
        self._c.set(cn.detach())

        return y


def test_packer():
    reset = torch.Tensor([
        [0, 1, 0, 0],
        [0, 0, 1, 0],
        [0, 0, 0, 1]]).bool().cuda()

    x = torch.Tensor([
        [0, 10, 20, 30],
        [1, 11, 40, 31],
        [2, 12, 41, 50]]).float().cuda()

    h = torch.Tensor([[1, 2, 3, 4]]).float().cuda()

    packer = Packer(reset)
    xp, (hp, cp) = packer.pack(x, h, h)
    xu, (hu, cu) = packer.unpack(xp, (hp, cp))

    torch.testing.assert_allclose(xp.data, torch.tensor([0, 10, 20, 40, 30, 50, 1, 11, 41, 31, 2, 12]).float().cuda())
    torch.testing.assert_allclose(xp.batch_sizes, torch.tensor([6, 4, 2]))
    torch.testing.assert_allclose(x, xu)

    torch.testing.assert_allclose(hp, torch.tensor([1, 0, 3, 0, 4, 0]).float().cuda())
    torch.testing.assert_allclose(hu, torch.tensor([[1, 0, 0, 0]]).float().cuda())




import torch
import numpy as np
from torch import nn
from torch.nn import functional as F
from rebar import recurrence
import matplotlib.pyplot as plt

class PositionalEmbedding(nn.Module):

    def __init__(self, d_model, lim=1024):
        super().__init__()
        assert d_model % 2 == 0

        self._lim = lim
        inv_freq = 2*np.pi / (lim ** (torch.arange(0.0, d_model, 2.0) / d_model))
        self.register_buffer('inv_freq', inv_freq)

    def forward(self, pos_seq):
        sinusoid_inp = pos_seq[..., None]*self.inv_freq
        return torch.cat([sinusoid_inp.sin(), sinusoid_inp.cos()], dim=-1)

    def _pattern(self, K, s):
        K = self._lim if K is None else K
        s = 1 if s is None else s
        pos_seq = torch.arange(0., K, s, dtype=torch.float, device=self.inv_freq.device)
        return self(pos_seq)

    def plot_pattern(self, K=None, s=None):
        plt.imshow(numpyify(self._pattern(K, s)))

    def plot_similarity(self, K=None, s=None):
        pattern = self._pattern(K, s)
        top = (pattern[None, :]*pattern[:, None]).sum(-1)
        bot = pattern.pow(2).sum(-1).pow(.5)
        plt.imshow(numpyify(top/(bot[None, :]*bot[:, None])), vmin=-1, vmax=+1, cmap=plt.cm.RdBu_r)
        plt.colorbar()

def rel_shift(x):
    """Explanation: https://github.com/kimiyoung/transformer-xl/issues/8#issuecomment-454458852"""
    (T, C), tail = x.shape[:2], x.shape[2:]

    zero_pad = x.new_zeros((T, 1) + tail)
    x_padded = torch.cat([zero_pad, x], dim=1)
    x_padded = x_padded.view((C+1, T) + tail)

    return x_padded[1:].view_as(x)

class ResetMasker(nn.Module):

    def __init__(self, mem_len):
        super().__init__()
        self._mem_len = mem_len
        self._reset = recurrence.State()

    @torch.no_grad()
    def __call__(self, reset):
        if reset.ndim == 1:
            return self(reset[None])
        old_reset = self._reset.get(lambda: torch.ones_like(reset[:0]))
        resets = torch.cat([old_reset, reset], 0)
        self._reset.set(resets[-self._mem_len:])

        diag_resets = torch.diag_embed(resets.T, offset=0, dim1=0, dim2=2)
        flipped = diag_resets.int().flip(2)
        rows = (flipped.cumsum(2) - flipped).flip(2)
        cols = (rows.cumsum(0)).bool()

        T = reset.size(0)
        return cols[-T:]

def attention_mask(T, M, mem_len, device=None, future=0):
    all_ones = torch.ones((T, M+T), dtype=torch.uint8, device=device)
    mask_len = M+T - mem_len
    mask_shift_len = (T - mask_len) if mask_len > 0 else T
    future_mask = torch.triu(all_ones, M+future)
    history_mask = torch.tril(all_ones, -mask_shift_len)
    default_mask = (future_mask + history_mask)[:, :, None].bool()
    return default_mask

class Weights(nn.Module):

    def __init__(self, mem_len, d_model, d_head=None, n_head=1, content=True, position=True, norm=True, memory=True):
        """Useful to be able to turn content/norm off for testing"""
        super().__init__()

        self.d_model = d_model
        self.mem_len = mem_len
        self.n_head = n_head
        self.d_head = (d_head or d_model//self.n_head)
        self.content = content
        self.position = position
        self.memory = memory

        self.m = recurrence.State()
        self.masker = ResetMasker(self.mem_len)

        self.norm = nn.LayerNorm(d_model) if norm else lambda x: x
        self.q = nn.Linear(d_model, self.n_head*self.d_head, bias=False)

        # Doesn't really matter how we initialize; LayerNorm will stop things from blowing up
        if self.content:
            self.k = nn.Linear(d_model, self.n_head*self.d_head, bias=False)
            self.k_bias = nn.Parameter(torch.empty((self.n_head, self.d_head)))
            torch.nn.init.normal_(self.k_bias, 0, 1)

        if self.position:
            self.pos_emb = PositionalEmbedding(self.d_model)
            self.r = nn.Linear(d_model, self.n_head*self.d_head, bias=False)
            self.r_bias = nn.Parameter(torch.empty((self.n_head, self.d_head)))
            torch.nn.init.normal_(self.r_bias, 0, 1)

    def forward(self, h, reset=None):
        NH, DH = self.n_head, self.d_head
        T, B = h.shape[:2]

        m = self.m.get(lambda: h.new_zeros((0, B, self.d_model)))
        if self.memory:
            self.m.set(torch.cat([m, h])[-self.mem_len:].detach())
        M = m.size(0)
        TM = T+M

        reset = torch.zeros((T, B), dtype=torch.bool, device=h.device) if reset is None else reset
        mask = self.masker(reset).permute(0, 2, 1)[:, -TM:]
        mask = mask | attention_mask(T, M, self.mem_len, device=h.device)
        mask = mask.unsqueeze(-1)

        cat = self.norm(torch.cat([m, h], 0))
        q = self.q(cat[-T:]).view(T, B, NH, DH)

        score = h.new_zeros((T, TM, B, NH))

        if self.content:
            k = self.k(cat).view(TM, B, NH, DH)
            attn_score = torch.einsum('ibnd,jbnd->ijbn', (q + self.k_bias, k))
            score += attn_score

        if self.position:
            pos_seq = torch.flip(torch.arange(TM, device=h.device, dtype=h.dtype), (0,))
            r = self.r(self.pos_emb(pos_seq)).view(TM, NH, DH)

            pos_score = torch.einsum('ibnd,jnd->ijbn', (q + self.r_bias, r))
            score += rel_shift(pos_score)

        #### compute attention probability
        score = (score
                        .div(self.d_head**.5)
                        .masked_fill(mask, -65000) # -65k is a very, very small number for a 16-bit float
                        .clamp(-65000, +65000)) # Clamp to suppress any infs that sometimes overflow when using 16-bit floats

        prob = F.softmax(score, dim=1)

        # Zero any output rows where the mask was all true
        return prob.where(~mask.all(1, keepdim=True), torch.zeros_like(prob))

class Values(nn.Module):

    def __init__(self, *args, **kwargs):
        super().__init__()

        self.weights = Weights(*args, **kwargs)
        self.d_model = self.weights.d_model
        self.n_head = self.weights.n_head
        self.d_head = self.weights.d_head
        self.mem_len = self.weights.mem_len
        
        self.norm = nn.LayerNorm(self.d_model)

        self.v = nn.Linear(self.d_model, self.n_head*self.d_head, bias=False)
        self.o = nn.Linear(self.n_head*self.d_head, self.d_model, bias=False)

    def forward(self, h, reset=None):
        NH, DH = self.n_head, self.d_head
        T, B = h.shape[:2]

        #TODO: Cache the qkv values rather than the h values?
        m = self.weights.m.get(lambda: h.new_zeros((0, B, self.d_model)))
        M = m.size(0)
        TM = T+M

        cat = self.norm(torch.cat([m, h], 0))
        v = self.v(cat).view(TM, B, NH, DH)

        weights = self.weights(h, reset)
        summary = torch.einsum('ijbn,jbnd->ibnd', weights, v)

        return F.relu(self.o(summary.reshape(T, B, NH*DH)))

class Gate(nn.Module):

    def __init__(self, d_model, bias=2.):
        super().__init__()
        self.W = nn.Linear(d_model, 3*d_model, bias=False)
        self.U = nn.Linear(d_model, 2*d_model, bias=False)
        self.Ug = nn.Linear(d_model, d_model, bias=False)
        self.b = nn.Parameter(torch.full((d_model,), bias))

    def forward(self, x, y):
        wr, wz, wg = torch.chunk(self.W(y), 3, dim=-1)
        ur, uz  = torch.chunk(self.U(y), 2, dim=-1)

        r = torch.sigmoid(wr + ur)
        z = torch.sigmoid(wz + uz - self.b)
        h = torch.tanh(wg + self.Ug(r*x))

        return (1 - z)*x + z*h

class GatedAttention(nn.Module):
    def __init__(self, **kwargs):
        super().__init__()

        self.attn = Values(**kwargs)
        self.attn_gate = Gate(self.attn.d_model)
        self.ff = nn.Sequential(
            nn.LayerNorm(self.attn.d_model),
            nn.Linear(self.attn.d_model, self.attn.d_model),
            nn.ReLU())
        self.ff_gate = Gate(self.attn.d_model)

    def forward(self, h, reset):
        a = self.attn(h, reset)
        h = self.attn_gate(h, a)
        return self.ff_gate(h, self.ff(h))

class Transformer(nn.Module):

    def __init__(self, n_layers=1, *args, **kwargs):
        super().__init__()
        self.layers = nn.ModuleList([GatedAttention(*args, **kwargs) for _ in range(n_layers)])

    def forward(self, h, reset):
        """h: TxBxC or BxC (will be treated as a single step)"""
        if h.ndim == 2:
            return self(h[None], reset[None])[0]

        for layer in self.layers:
            h = layer(h, reset)
        return h


def test_weights_simple():
    T, B, K = 5, 7, 2
    h = torch.rand((T, B, K))
    attn = Weights(T, K)
    attn(h)
    attn(h)

def test_weights_content_single():
    h = torch.tensor([10, 0, 10])[:, None, None].float()

    attn = Weights(3, 1, position=False, norm=False)
    attn.k.weight[:] = +1.
    attn.q.weight[:] = +1.
    attn.k_bias[:] = 0.
    what = attn(h)

    w = torch.tensor([[0, 0, 0], [1, 0, 0], [1, 0, 0]]).float()
    torch.testing.assert_allclose(w, what[:, :, 0, 0])

def test_weights_content_multi():
    h0 = torch.tensor([10, 0])[:, None, None].float()
    h1 = torch.tensor([10,])[:, None, None].float()

    attn = Weights(3, 1, position=False, norm=False)
    attn.k.weight[:] = +1.
    attn.q.weight[:] = +1.
    attn.k_bias[:] = 0.

    attn(h0)
    what = attn(h1)

    w = torch.tensor([[1, 0, 0]]).float()
    torch.testing.assert_allclose(w, what[:, :, 0, 0])

def test_values_simple():
    T, B, K = 5, 7, 2
    h = torch.rand((T, B, K))
    attn = Values(T, K)
    attn(h)
    attn(h)

def test_reset():
    h = torch.rand((8, 1, 2))
    reset = torch.zeros((8, 1), dtype=torch.bool)
    reset[6] = True

    weights = Weights(4, 2)
    weights(h, reset)
    attn = weights(h, reset)

    assert (attn[0:, :2] == 0).all()
    assert (attn[6:, :10] == 0).all()


from collections import OrderedDict
from functools import wraps

SCREEN_WIDTH = 119
SCREEN_HEIGHT = 200

class dotdict(OrderedDict):
    """dotdicts are dictionaries with additional support for attribute (dot) access of their elements.
    dotdicts have a lot of unusual but extremely useful behaviours, which are documented in :ref:`the dotdicts
    and arrdicts concept section <dotdicts>` .

    """
    
    def __dir__(self):
        return sorted(set(super().__dir__() + list(self.keys())))

    def __getattr__(self, key):
        if key in self:
            return self[key]
        else:
            try:
                return type(self)([(k, getattr(v, key)) for k, v in self.items()])
            except AttributeError:
                raise AttributeError(f"There is no member called '{key}' and one of the leaves has no attribute '{key}'") from None

    def __call__(self, *args, **kwargs):
        return type(self)([(k, v(*args, **kwargs)) for k, v in self.items()])

    def __str__(self):
        return treestr(self)
    
    def __repr__(self):
        return self.__str__()

    def __getstate__(self):
        return self

    def __setstate__(self, state):
        self.update(state)
    
    def copy(self):
        """Shallow-copy the dotdict"""
        return type(self)(super().copy()) 
    
    def pipe(self, f, *args, **kwargs):
        """Returns ``f(self, *args, **kwargs)`` . 

        >>> d = dotdict(a=1, b=2)
        >>> d.pipe(list)
        ['a', 'b']

        Useful for method-chaining."""
        return f(self, *args, **kwargs)

    def map(self, f, *args, **kwargs):
        """Applies ``f`` to the values of the dotdict, returning a matching dotdict of the results.
        ``*args`` and  ``**kwargs`` are passed as extra arguments to each call.

        >>> d = dotdict(a=1, b=2)
        >>> d.map(int.__add__, 10)
        dotdict:
        a    11
        b    12

        Useful for method-chaining. Works equally well on trees of dotdicts.
        
        See :func:`mapping` for a functional version of this method."""
        return mapping(f)(self, *args, **kwargs)

    def starmap(self, f, *args, **kwargs):
        """Applies ``f`` to the values of the dotdicts one key at a time, returning a matching dotdict of the results.

        >>> d = dotdict(a=1, b=2)
        >>> d.starmap(int.__add__, d)
        dotdict:
        a    2
        b    4

        Useful for method-chaining. Works equally well on trees of dotdicts.
        
        See :func:`starmapping` for a functional version of this method."""
        return starmapping(f)(self, *args, **kwargs)

def treestr(t):
    """Stringifies a tree structure. These turn up all over the place in my code, so it's worth factoring out"""
    key_length = max(map(len, map(str, t.keys()))) if t.keys() else 0
    max_spaces = 4 + key_length
    val_length = SCREEN_WIDTH - max_spaces
    
    d = {}
    for k, v in t.items():
        if isinstance(v, dotdict):
            d[k] = str(v)
        elif isinstance(v, (list, set, dict)):
            d[k] = f'{type(v).__name__}({len(v)},)'
        elif hasattr(v, 'shape') and hasattr(v, 'dtype'):                    
            d[k] = f'{type(v).__name__}({tuple(v.shape)}, {v.dtype})'
        elif hasattr(v, 'shape'):
            d[k] = f'{type(v).__name__}({tuple(v.shape)})'
        else:
            lines = str(v).splitlines()
            if (len(lines) > 1) or (len(lines[0]) > val_length):
                d[k] = lines[0][:val_length] + ' ...'
            else:
                d[k] = lines[0]

    s = [f'{type(t).__name__}:']
    for k, v in d.items():
        lines = v.splitlines() or ['']
        s.append(str(k) + ' '*(max_spaces - len(str(k))) + lines[0])
        for l in lines[1:]:
            s.append(' '*max_spaces + l)
        if len(s) >= SCREEN_HEIGHT-1:
            s.append('...')
            break

    return '\n'.join(s)

def mapping(f):
    """Wraps ``f`` so that when called on a dotdict, ``f`` instead gets called on the dotdict's values
    and a dotdict of the results is returned. Extra ``*args`` and ``**kwargs`` passed to the wrapper are
    passed as extra arguments to ``f`` .

    >>> d = dotdict(a=1, b=2)
    >>> m = mapping(int.__add__)
    >>> m(d, 10)
    dotdict:
    a    11
    b    12
    
    Works equally well on trees of dotdicts, where ``f`` will be applied to the leaves of the tree.

    Can be used as a decorator.

    See :func:`dotdict.map` for an object-oriented version of this function.
    """

    @wraps(f)
    def g(x, *args, **kwargs):
        if isinstance(x, dict):
            return type(x)([(k, g(v, *args, **kwargs)) for k, v in x.items()])
        if isinstance(f, str):
            return getattr(x, f)(*args, **kwargs)
        return f(x, *args, **kwargs)
    return g

def starmapping(f):
    """Wraps ``f`` so that when called on a sequence of dotdicts, ``f`` instead gets called on the dotdict's values
    and a dotdict of the results is returned.

    >>> d = dotdict(a=1, b=2)
    >>> m = starmapping(int.__add__)
    >>> m(d, d)
    dotdict:
    a    2
    b    4
    
    Works equally well on trees of dotdicts, where ``f`` will be applied to the leaves of the trees.

    Can be used as a decorator.

    See :func:`dotdict.starmap` for an object-oriented version of this function.
    """
    @wraps(f)
    def g(x, *args, **kwargs):
        if isinstance(x, dict):
            return type(x)([(k, g(x[k], *(a[k] for a in args))) for k in x])
        if isinstance(f, str):
            return getattr(x, f)(*args)
        else:
            return f(x, *args)
    return g

def leaves(t):
    """Returns the leaves of a tree of dotdicts as a list"""
    if isinstance(t, dict):
        return [l for v in t.values() for l in leaves(v)]
    return [t]


"""**rebar** helps with reinforcement. That's why it's called rebar! It's a toolkit that has evolved
as I've worked on RL projects.

Unlike the :mod:`megastep` module which is stable, documented and feature-complete, rebar is an unstable,
undocumented work-in-progress. It's in the megastep repo because megastep itself uses two of rebar's most useful components:
:class:`~rebar.dotdict.dotdict` and :class:`~rebar.arrdict.arrdict`, while the demo uses a whole lot more.
"""


import torch
import time
from contextlib import contextmanager
from functools import partial
from torch import nn
import logging
import pandas as pd
from .. import paths

log = logging.getLogger(__name__)

# For re-export
from .writing import *
from .writing import to_dir, mean
from .reading import from_dir, Reader
from . import gpu

@contextmanager
def via_dir(run_name, *args, **kwargs):
    with to_dir(run_name), from_dir(run_name, *args, **kwargs):
        yield

def normhook(name, t):

    def hook(grad):
        mean(name, grad.pow(2).sum().pow(.5))

    t.register_hook(hook)

def total_gradient_norm(params):
    if isinstance(params, nn.Module):
        return total_gradient_norm(params.parameters())
    norms = [p.grad.data.float().pow(2).sum() for p in params if p.grad is not None]
    return torch.sum(torch.tensor(norms)).pow(.5)

def total_norm(params):
    if isinstance(params, nn.Module):
        return total_norm(params.parameters())
    return sum([p.data.float().pow(2).sum() for p in params if p is not None]).pow(.5)

def rel_gradient_norm(name, agent):
    mean(name, total_gradient_norm(agent), total_norm(agent))

def funcduty(name):
    def factory(f):
        def g(self, *args, **kwargs):
            start = time.time()
            result = f(self, *args, **kwargs)
            record('duty', f'duty/{name}', time.time() - start)
            return result
        return g
    return factory

def compare(run_names=[-1], prefix='', rule='60s'):
    return pd.concat({paths.resolve(run): Reader(run, prefix).resample(rule) for run in run_names}, 1)

## TESTS

def test_from_dir():
    from .. import paths, widgets, logging
    paths.clear('test-run', 'stats')
    paths.clear('test-run', 'logs')

    compositor = widgets.Compositor()
    with logging.from_dir('test-run', compositor), \
            to_dir('test-run'), \
            from_dir('test-run', compositor):
        for i in range(10):
            mean('count', i)
            mean('twocount', 2*i)
            time.sleep(.25) 



"""TODO-DOCS Deathmatch docs"""
import torch
from megastep import modules, core, plotting, spaces, scene, cubicasa
from rebar import arrdict, dotdict
import matplotlib.pyplot as plt
import numpy as np
import matplotlib as mpl

CLEARANCE = 1.

@dotdict.mapping
def expand(x):
    B, A = x.shape[:2]
    return x.reshape(B*A, 1, *x.shape[2:])

@dotdict.mapping
def collapse(x, n_agents):
    B = x.shape[0]
    return x.reshape(B//n_agents, n_agents, *x.shape[2:])

class Deathmatch:

    def __init__(self, n_envs, n_agents, *args, **kwargs):
        geometries = cubicasa.sample(max(n_envs//4, 1))
        scenery = scene.scenery(geometries, n_agents)
        self.core = core.Core(scenery, *args, res=4*128, fov=70, **kwargs)
        self._rgb = modules.RGB(self.core, n_agents=1, subsample=4)
        self._depth = modules.Depth(self.core, n_agents=1, subsample=4)
        self._imu = modules.IMU(self.core, n_agents=1)
        self._movement = modules.MomentumMovement(self.core, n_agents=1)
        self._spawner = modules.RandomSpawns(geometries, self.core)

        self.action_space = self._movement.space
        self.obs_space = dotdict.dotdict(
            rgb=self._rgb.space,
            d=self._depth.space,
            imu=self._imu.space,
            health=spaces.MultiVector(1, 1))

        self._bounds = arrdict.torchify(np.stack([g.masks.shape*g.res for g in geometries])).to(self.core.device)
        self._health = self.core.agent_full(np.nan)
        self._damage = self.core.agent_full(np.nan)

        self.n_envs = self.core.n_envs*self.core.n_agents
        self.device = self.core.device

    def _reset(self, reset=None):
        reset = (self._health <= 0) if reset is None else reset
        self._spawner(reset)
        self._health[reset] = 1.
        self._damage[reset] = 0.
        return reset.reshape(-1)

    def _shoot(self, opponents):
        res = opponents.size(-1)
        middle = slice(res//2-1, res//2+1)
        agents = torch.arange(self.core.n_agents, device=self.core.device)
        matchings = (opponents[:, :, None] == agents[None, None, :, None, None])[..., middle].any(-1).any(-1)
        self.matchings = matchings
        
        hits = matchings.sum(2).float()
        wounds = matchings.sum(1).float()

        self._damage[:] += .05*hits

        pos = self.core.agents.positions 
        outside = (pos < -CLEARANCE).any(-1) | (pos > (self._bounds[:, None] + CLEARANCE)).any(-1)

        # 5% damage per hit, .1% damage per timestep
        self._health[:] += -.05*(wounds + outside) - .001
        
        return hits.reshape(-1)

    def _observe(self):
        r = modules.render(self.core)
        line_idxs = modules.downsample(r.indices, self._rgb.subsample)[..., self._rgb.subsample//2]
        obj_idxs = line_idxs//len(self.core.scenery.model)
        mask = (0 <= line_idxs) & (obj_idxs < self.core.n_agents)
        opponents = obj_idxs.where(mask, torch.full_like(line_idxs, -1))
        hits = self._shoot(opponents)
        obs = arrdict.arrdict(
                rgb=self._rgb(r), 
                d=self._depth(r), 
                imu=self._imu(),
                health=self._health.unsqueeze(-1).clone())
        return obs, hits

    @torch.no_grad()
    def reset(self):
        reset = self._reset(self.core.agent_full(True))
        obs, reward = self._observe()
        return arrdict.arrdict(
            obs=expand(obs),
            reward=reward,
            reset=reset)

    @torch.no_grad()
    def step(self, decision):
        reset = self._reset()
        self._movement(collapse(decision, self.core.n_agents))
        obs, reward = self._observe()
        return arrdict.arrdict(
            obs=expand(obs),
            reward=reward,
            reset=reset)

    def state(self, e=0):
        return arrdict.arrdict(
            core=self.core.state(e),
            rgb=self._rgb.state(e),
            d=self._depth.state(e),
            health=self._health[e].clone(),
            damage=self._damage[e].clone(),
            matchings=self.matchings[e].clone(),
            bounds=self._bounds[e].clone())

    @classmethod
    def plot_state(cls, state):
        n_agents = state.core.n_agents
        show_value = 'decision' in state

        fig = plt.figure()
        gs = plt.GridSpec(n_agents, 4 if show_value else 3, fig)

        colors = [f'C{i}' for i in range(n_agents)]

        plan = core.Core.plot_state(state.core, plt.subplot(gs[:-1, :-1]))

        # Add hits
        origin, dest = state.matchings.nonzero()
        lines = state.core.agents.positions[np.stack([origin, dest], 1)]
        linecolors = np.array(colors)[origin]
        lines = mpl.collections.LineCollection(lines, color=linecolors, linewidth=1, alpha=.5)
        plan.add_collection(lines)

        # Add bounding box
        size = state.bounds[::-1] + 2*CLEARANCE
        bounds = mpl.patches.Rectangle(
            (-CLEARANCE, -CLEARANCE), *size, 
            linewidth=1, edgecolor='k', facecolor=(0., 0., 0., 0.))
        plan.add_artist(bounds)

        # Add observations
        images = {'rgb': state.rgb, 'd': state.d}
        plotting.plot_images(images, [plt.subplot(gs[i, -1]) for i in range(n_agents)])

        ax = plt.subplot(gs[-1, 0])
        ax.barh(np.arange(n_agents), state.health, color=colors)
        ax.set_ylabel('health')
        ax.set_yticks([])
        ax.invert_yaxis()
        ax.set_xlim(0, 1)

        ax = plt.subplot(gs[-1, 1])
        ax.barh(np.arange(n_agents), state.damage, color=colors)
        ax.set_ylabel('inflicted')
        ax.set_yticks([])
        ax.invert_yaxis()

        if show_value:
            ax = plt.subplot(gs[-1, 2])
            ax.barh(np.arange(n_agents), state.decision.value, color=colors)
            ax.set_ylabel('value')
            ax.set_yticks([])
            ax.invert_yaxis()

        return fig

    def display(self, e=0):
        return self.plot_state(arrdict.numpyify(self.state(e=e)))


import numpy as np
import time
import torch
import matplotlib.pyplot as plt
import pandas as pd
import logging

log = logging.getLogger(__name__)

CATEGORIES = {}
def category(M):
    CATEGORIES[M.__name__.lower()] = M
    return M

@category
def last(x):
    def resample(**kwargs):
        return x.resample(**kwargs).last()
    return resample

@category
def max(x):
    def resample(**kwargs):
        return x.resample(**kwargs).max()
    return resample

@category
def mean(total, count=1):
    def resample(**kwargs):
        return total.resample(**kwargs).mean()/count.resample(**kwargs).mean()
    return resample

@category
def std(x):
    def resample(**kwargs):
        return x.resample(**kwargs).std()
    return resample

@category
def cumsum(total=1):
    def resample(**kwargs):
        return total.resample(**kwargs).sum().cumsum()
    return resample

@category
def timeaverage(x):
    def resample(**kwargs):
        # TODO: To do this properly, I need to get individual per-device streams
        y = x.sort_index()
        dt = y.index.to_series().diff().dt.total_seconds()
        return (y*dt).resample(**kwargs).mean()/dt.resample(**kwargs).mean()
    return resample

@category
def duty(duration):
    def resample(**kwargs):
        sums = duration.resample(**kwargs).sum()
        periods = sums.index.to_series().diff().dt.total_seconds()
        return sums/periods
    return resample

@category
def maxrate(duration, count=1):
    def resample(**kwargs):
        return count.resample(**kwargs).mean()/duration.resample(**kwargs).mean()
    return resample

@category
def rate(count=1):
    def resample(**kwargs):
        counts = count.resample(**kwargs).sum()
        dt = pd.to_timedelta(counts.index.freq).total_seconds()
        dt = min(dt, (count.index[-1] - count.index[0]).total_seconds())
        return counts/dt
    return resample

@category
def period(count=1):
    def resample(**kwargs):
        counts = count.resample(**kwargs).sum()
        dt = pd.to_timedelta(counts.index.freq).total_seconds()
        dt = min(dt, (count.index[-1] - count.index[0]).total_seconds())
        return dt/counts
    return resample

@category
def dist(samples, size=10000):
    return samples

@category
def noisescale(S, G2, B):
    def resample(**kwargs):
        return S.resample(**kwargs).mean()/G2.resample(**kwargs).mean()
    return resample


import numpy as np
import torch
from rebar import arrdict, dotdict
import pandas as pd

__all__ = ['dataframe']

class MultiVector:

    def __init__(self, n_agents, dim):
        super().__init__()
        self.shape = (n_agents, dim)

class MultiDiscrete:

    def __init__(self, n_agents, n_actions):
        super().__init__()
        self.shape = (n_agents, n_actions)

def _dataframe(traj):
    if isinstance(traj, dict):
        return [([k] + kk, vv) for k, v in traj.items() for kk, vv in _dataframe(v)]
    if isinstance(traj, torch.Tensor):
        return [([], pd.Series([x for x in arrdict.numpyify(traj)]))]

def dataframe(traj, **kwargs):
    return pd.concat({'.'.join(k): v for k, v in _dataframe({**traj, **kwargs})}, 1)

class FSM:

    def __init__(self, n_envs, fsm, device='cuda'):
        self.n_envs = n_envs
        self.n_states = fsm.n_states
        self.device = torch.device(device)

        self._obs = fsm.obs.to(device)
        self._trans = fsm.trans.to(device)
        self._reward = fsm.reward.to(device)
        self._terminal = fsm.terminal.to(device)
        self._start = fsm.start.to(device)
        self._indices = fsm.indices
        self._names = fsm.names

        self._token = torch.full((self.n_envs,), -1, dtype=torch.long, device=device)

        self.obs_space = MultiVector(1, fsm.d_obs) if fsm.d_obs else spaces.MultiEmpty()
        self.action_space = MultiDiscrete(1, fsm.n_actions)

    def _reset(self, reset):
        if reset.any():
            n_reset = reset.sum()
            self._token[reset] = torch.distributions.Categorical(self._start).sample((n_reset,))

    def reset(self):
        self._reset(self._terminal.new_ones((self.n_envs,)))
        return arrdict.arrdict(
            obs=self._obs[self._token, None],
            reward=torch.zeros((self.n_envs,), dtype=torch.float, device=self.device),
            reset=torch.ones((self.n_envs), dtype=torch.bool, device=self.device),
            terminal=torch.ones((self.n_envs), dtype=torch.bool, device=self.device))

    def step(self, decision):
        actions = decision.actions[:, 0]
        reward = self._reward[self._token, actions]

        weights = self._trans[self._token, actions]
        self._token[:] = torch.distributions.Categorical(weights).sample()
        
        reset = self._terminal[self._token]
        self._reset(reset)

        return arrdict.arrdict(
            obs=self._obs[self._token, None],
            idx=self._token.clone(),
            reward=reward,
            reset=reset,
            terminal=reset)

    def solve(self, eps=1e-3, gamma=.99):
        value = self._reward.new_zeros((self.n_states,))
        while True:
            succ = (value[None, None, :]*self._trans).sum(-1)
            best = (self._reward + gamma*succ).max(-1)
            best.values[self._terminal] = 0
            change = value - best.values
            value = best.values
            
            if change.pow(2).mean().pow(.5) < eps:
                break
                
        return arrdict.arrdict(value=value, policy=best.indices)

    def dataframe(self, **kwargs):
        soln = self.solve(**kwargs)
        successor = self._trans[torch.arange(self.n_states, device=self.device), soln.policy].argmax(-1)
        successor = [self._names[i] for i in successor]
        df = pd.DataFrame(arrdict.numpyify(dict(
                    name=self._names,
                    obs=[tuple(f'{x:.2f}' for x in o) for o in arrdict.numpyify(self._obs)],
                    term=self._terminal,
                    start=self._start,
                    value=soln.value,
                    policy=soln.policy,
                    successor=successor,
                ))).sort_index()
        df.index.name = 'idx'
        return df

    def __repr__(self):
        s, a, _ = self._trans.shape
        return f'{type(self).__name__}({s}s{a}a)' 

    def __str__(self):
        return repr(self)


class State:

    def __init__(self, name, builder):
        self._name = name
        self._builder = builder

    def to(self, state, action=0, reward=0., weight=1.):
        action = int(action)
        self._builder._trans.append(dotdict.dotdict(
            prev=self._name, 
            action=action, 
            next=state, 
            reward=reward, 
            weight=weight))
        return self

    def state(self, *args, **kwargs):
        return self._builder.state(*args, **kwargs)

    def build(self):
        return self._builder.build()

class Builder:

    def __init__(self):
        self._obs = []
        self._trans = []

    def state(self, name, obs, start=0.):
        if isinstance(obs, (int, float, bool)):
            obs = (obs,)
        self._obs.append(dotdict.dotdict(state=name, obs=obs, start=start))
        return State(name, self)
    
    def build(self):
        states = (
            {x.state for x in self._obs} | 
            {x.prev for x in self._trans} | 
            {x.next for x in self._trans})

        actions = {x.action for x in self._trans}
        assert max(actions) == len(actions)-1, 'Action set isn\'t contiguous'
        
        indices = {s: i for i, s in enumerate(states)}
        names = np.array(list(states))

        n_states = len(states)
        n_actions = len(actions)
        (d_obs,) = {len(x.obs) for x in self._obs}

        obs = torch.full((n_states, d_obs), np.nan)
        start = torch.full((n_states,), 0.)
        for x in self._obs:
            obs[indices[x.state]] = torch.as_tensor(x.obs)
            start[indices[x.state]] = x.start

        trans = torch.full((n_states, n_actions, n_states), 0.)
        reward = torch.full((n_states, n_actions), 0.)
        for x in self._trans:
            trans[indices[x.prev], x.action, indices[x.next]] = x.weight
            reward[indices[x.prev], x.action] = x.reward
        
        terminal = (trans.sum(-1).max(-1).values == 0)

        assert start.sum() > 0, 'No start state declared'

        return dotdict.dotdict(
            obs=obs, trans=trans, reward=reward, terminal=terminal, start=start, 
            indices=indices, names=names,
            n_states=n_states, n_actions=n_actions, d_obs=d_obs)


def fsm(f):

    def init(self, n_envs=1, *args, **kwargs):
        fsm = f(*args, **kwargs)
        assert isinstance(fsm, dict), 'FSM description must be a dictionary. Did you forget to call `.build()`?'
        super(self.__class__, self).__init__(n_envs, fsm)

    name = f.__name__
    __all__.append(name)
    return type(name, (FSM,), {'__init__': init})

@fsm
def ObliviousConstantReward():
    return (Builder()
        .state('start', obs=(), start=1.)
            .to('end', reward=1.)
        .build())

@fsm
def ObliviousCyclicReward():
    return (Builder()
        .state('start', obs=0., start=1.)
            .to('middle', reward=1)
        .state('middle', obs=1.)
            .to('end', reward=0)
        .build())

@fsm
def ObliviousChain(n=2, r=1):
    assert n >= 2, 'Need the number of states to be at least 2'
    b = Builder()
    b.state(0, obs=0., start=1.).to(1, 0)
    for i in range(1, n):
        reward = (i == n-1)
        b.state(i, obs=i/n).to(i+1, 0, reward=reward)
    return b.build()

@fsm
def ObliviousCoin():
    return (Builder()
        .state('heads', obs=+1., start=1.)
            .to('end', 0, reward=+1)
        .state('tails', obs=-1., start=1.)
            .to('end', 0, reward=-1)
        .build())

@fsm
def ObliviousDelayedCoin():
    return (Builder()
        .state('heads-1', obs=+.5, start=1.)
            .to('heads-2')
        .state('heads-2', obs=+1.)
            .to('end', reward=+1)
        .state('tails-1', obs=-.5, start=1.)
            .to('tails-2')
        .state('tails-2', obs=-1.)
            .to('end', reward=-1)
        .build())

@fsm
def DelayedMatchCoin():
    return (Builder()
        .state('heads-1', obs=+1., start=1.)
            .to('heads-2', 0)
            .to('heads-2', 1)
        .state('heads-2', obs=0.)
            .to('end', 0, reward=+1)
            .to('end', 1, reward=-1)
        .state('tails-1', obs=0., start=1.)
            .to('tails-2', 0)
            .to('tails-2', 1)
        .state('tails-2', obs=-1.)
            .to('end', 0, reward=-1)
            .to('end', 1, reward=+1)
        .build())

@fsm
def MatchCoin():
    return (Builder()
        .state('heads', obs=+1., start=1.)
            .to('end', 0, reward=+1)
            .to('end', 1, reward=-1)
        .state('tails', obs=-1., start=1.)
            .to('end', 0, reward=-1)
            .to('end', 1, reward=+1)
        .build())

@fsm
def RandomChain(n=2, seed=0):
    assert n >= 2, 'Need the radius to be at least 2'
    b = Builder()
    random = np.random.RandomState(seed)
    actions = random.permutation([0, 1])
    (b.state(0, obs=0., start=1.)
        .to(0, action=actions[0])
        .to(1, action=actions[1]))
    for i in range(1, n):
        reward = (i == n-1)
        actions = random.permutation([0, 1])
        (b.state(+i, obs=+i/n)
            .to(0, action=actions[0])
            .to(i+1, action=actions[1], reward=+reward))
    return b.build()

  


#!/usr/bin/env python

from distutils.core import setup
from setuptools import find_packages

setup(
    name='megastep',
    version='0.2.3',
    description='Helps build million-frame-per-second reinforcement learning environments',
    author='Andy Jones',
    author_email='andyjones.ed@gmail.com',
    url='http://andyljones.com/megastep',
    packages=find_packages(include=['megastep*', 'rebar*']),
    python_requires='>=3.6',
    install_requires=[
        'numpy>=1.18',
        'torch>=1.5',
        'torchvision>=0.6',
        'tqdm>=4',
        'matplotlib>=3',
        'ninja>=1.10'],
    extras_require={
        'cubicasa': [
            'beautifulsoup4>=4', 
            'shapely>=1.7', 
            'rasterio>=1.1',
            'pandas>=1'],
        'rebar': [
            'av>=8',
            'bokeh>=2',
            'ipywidgets>=7',
            'psutil>=5',
            'pandas>=1'],
        'docs': [
            'sphinx>=3'],
        'test': [
            'pytest>=5']},
    package_data={'megastep': ['src/*']}
)


"""TODO-DOCS Spaces docs"""

class MultiEmpty:
    
    def __init__(self):
        pass

class MultiVector:

    def __init__(self, n_agents, dim):
        super().__init__()
        self.shape = (n_agents, dim)

class MultiImage:

    def __init__(self, n_agents, C, H, W):
        super().__init__()
        self.shape = (n_agents, C, H, W)

class MultiConstant:

    def __init__(self, n_agents):
        self.shape = (n_agents,)

class MultiDiscrete:

    def __init__(self, n_agents, n_actions):
        super().__init__()
        self.shape = (n_agents, n_actions)


"""TODO-DOCS Plotting docs"""
import torch
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from . import core
from rebar import arrdict, dotdict
from matplotlib import tight_bbox

VIEW_RADIUS = 5

def imshow_arrays(arrs, transpose=False):
    """Args:
        arrs: `{name: A x C x H x W}`
    """
    arrs = {k: v.transpose(0, 3, 1, 2) if transpose else v for k, v in arrs.items()}
    [A] = {v.shape[0] for v in arrs.values()}
    ims = {}
    for a in range(A):
        layers = []
        for k, v in arrs.items():
            layer = v[a].astype(float)
            if layer.shape[0] == 1:
                layer = layer.repeat(3, 0)
            else:
                layer = core.gamma_encode(layer)
            layers.append(layer)
        layers = np.concatenate(layers, 1)
        ims[a] = layers.transpose(1, 2, 0)
    return ims

def plot_images(arrs, axes=None, aspect=1, **kwargs):
    ims = imshow_arrays(arrs, **kwargs)
    A = len(ims)
    H, W = ims[0].shape[:2]

    axes = plt.subplots(A, 1, squeeze=False)[1].flatten() if axes is None else axes
    
    # Aspect is height/width
    aspect = aspect/min(A, 4)*W/H
    for a in range(A):
        ax = axes[a]
        ax.imshow(ims[a], aspect=aspect, interpolation='none')
        ax.set_yticks(np.arange(H))
        ax.set_ylim(H-.5, -.5)
        ax.set_yticklabels(arrs.keys())
        ax.set_xticks([])
        ax.set_title(f'agent #{a}', fontdict={'color': f'C{a}', 'weight': 'bold'})

    return axes

def n_agent_texels(scenery):
    A = scenery.n_agents
    M = len(scenery.model)
    return scenery.textures.widths[:A*M].sum()

def line_arrays(state):
    scenery = state.scenery

    starts = scenery.textures.starts
    tex_starts = np.zeros(len(scenery.textures.vals), dtype=int)
    tex_starts[starts] = 1
    tex_starts = tex_starts.cumsum() - 1
    tex_offsets = np.arange(len(tex_starts)) - starts[tex_starts]

    seg_starts = tex_offsets/scenery.textures.widths[tex_starts]
    seg_starts = scenery.lines[tex_starts, 0]*(1 - seg_starts[:, None]) + scenery.lines[tex_starts, 1]*seg_starts[:, None]

    seg_ends = (tex_offsets+1)/scenery.textures.widths[tex_starts]
    seg_ends = scenery.lines[tex_starts, 0]*(1 - seg_ends[:, None]) + scenery.lines[tex_starts, 1]*seg_ends[:, None]

    lines = np.stack([seg_starts, seg_ends]).transpose(1, 0, 2)

    baked = scenery.baked.vals.copy()
    baked[:n_agent_texels(scenery)] = 1.

    colors = core.gamma_encode(scenery.textures.vals*baked[:, None])
    return lines, colors

def plot_lights(ax, state):
    vmin = state.scenery.lights[:, 2].min() - 1e-2
    vmax = state.scenery.lights[:, 2].max()
    for light in state.scenery.lights:
        alpha = (light[2] - vmin)/(vmax - vmin)
        ax.add_patch(mpl.patches.Circle(light[:2], radius=.05, alpha=alpha, color='yellow'))

def extent(state, zoom, radius=VIEW_RADIUS):
    if zoom:
        r, t = state.agents.positions.max(0) + radius 
        l, b = state.agents.positions.min(0) - radius
    else:
        r, t = state.scenery.lines.max(0).max(0) + 1
        l, b = state.scenery.lines.min(0).min(0) - 1

    w = max(t - b, r - l)/2
    cx, cy = (r + l)/2, (t + b)/2

    return (cx-w, cx+w), (cy-w, cy+w)

def plot_lines(ax, state, zoom=True):
    lines, colors = line_arrays(state)

    (l, r), (b, t) = extent(state, zoom)
    xs, ys = lines[:, :, 0], lines[:, :, 1]
    mask = (l < xs) & (xs < r) & (b < ys) & (ys < t)
    mask = mask.any(-1)

    seen = mpl.collections.LineCollection(lines[mask], colors=colors[mask], linestyle='solid', linewidth=2)
    ax.add_collection(seen)

def adjust_view(ax, state, zoom=True):
    xs, ys = extent(state, zoom)

    ax.set_xlim(*xs)
    ax.set_ylim(*ys)

    ax.set_aspect(1)
    ax.set_facecolor('#c6c1b3')

def plot_wedge(ax, pose, distance, fov, radians=False, **kwargs):
    scale = 180/np.pi if radians else 1
    left = scale*pose.angles - fov/2
    right = scale*pose.angles + fov/2
    width = distance - core.AGENT_RADIUS 
    wedge = mpl.patches.Wedge(
                    pose.positions, distance, left, right, width=width, **kwargs)
    ax.add_patch(wedge)

def plot_fov(ax, state, distance=1, field='agents'):
    a = len(getattr(state, field).angles)
    for i in range(a):
        plot_wedge(ax, getattr(state, field)[i], distance, state.fov, color=f'C{i}', alpha=.1)

def plot_poses(poses, ax=None, radians=True, color='C9', **kwargs):
    """Not used directly here, but often useful for code using this module"""
    ax = ax or plt.subplot()
    for angle, position in zip(poses.angles, poses.positions):
        ax.add_patch(mpl.patches.Circle(position, radius=core.AGENT_RADIUS, edgecolor=color, facecolor=[0,0,0,0]))

        scale = 1 if radians else np.pi/180
        offset = core.AGENT_RADIUS*np.array([np.cos(scale*angle), np.sin(scale*angle)])
        line = np.stack([position, position + offset])
        ax.plot(*line.T, color=color)
    return ax



from pkg_resources import resource_filename
import sysconfig
import torch.utils.cpp_extension

DEBUG = False

def _cuda():
    [torch_libdir] = torch.utils.cpp_extension.library_paths()
    python_libdir = sysconfig.get_config_var('LIBDIR')
    libpython_ver = sysconfig.get_config_var('LDVERSION')
    return torch.utils.cpp_extension.load(
        name='megastepcuda', 
        sources=[resource_filename(__package__, f'src/{fn}') for fn in ('wrappers.cpp', 'kernels.cu')], 
        extra_cflags=['-std=c++17'] + (['-g'] if DEBUG else []), 
        extra_cuda_cflags=['--use_fast_math', '-lineinfo', '-std=c++14'] + (['-g', '-G'] if DEBUG else []),
        extra_ldflags=[
            f'-lpython{libpython_ver}', '-ltorch', '-ltorch_python', '-lc10_cuda', '-lc10', 
            f'-L{torch_libdir}', f'-Wl,-rpath,{torch_libdir}',
            f'-L{python_libdir}', f'-Wl,-rpath,{python_libdir}'])
cuda = _cuda()



import torch
import pandas as pd
from io import BytesIO
from subprocess import check_output
from . import writing
import time


def memory(device=0):
    total_mem = torch.cuda.get_device_properties(f'cuda:{device}').total_memory
    writing.max(f'gpu-memory/cache/{device}', torch.cuda.max_memory_cached(device)/total_mem)
    torch.cuda.reset_max_memory_cached()
    writing.max(f'gpu-memory/alloc/{device}', torch.cuda.max_memory_allocated(device)/total_mem)
    torch.cuda.reset_max_memory_allocated()
    torch.cuda.reset_max_memory_cached()

def dataframe():
    """Use `nvidia-smi --help-query-gpu` to get a list of query params"""
    params = {
        'device': 'index', 
        'compute': 'utilization.gpu', 'access': 'utilization.memory', 
        'memused': 'memory.used', 'memtotal': 'memory.total',
        'fan': 'fan.speed', 'power': 'power.draw', 'temp': 'temperature.gpu'}
    command = f"""nvidia-smi --format=csv,nounits,noheader --query-gpu={','.join(params.values())}"""
    df = pd.read_csv(BytesIO(check_output(command, shell=True)), header=None)
    df.columns = list(params.keys())
    df = df.set_index('device')
    df = df.apply(pd.to_numeric, errors='coerce')
    return df

_last = -1
def vitals(device=None, throttle=0):
    # This is a fairly expensive op, so let's avoid doing it too often
    global _last
    if time.time() - _last < throttle:
        return
    _last = time.time()

    df = dataframe()
    if device is None:
        pass
    elif isinstance(device, int):
        df = df.loc[[device]]
    else:
        df = df.loc[device]

    fields = ['compute', 'access', 'fan', 'power', 'temp']
    for (device, field), value in df[fields].stack().iteritems():
        writing.mean(f'gpu/{field}/{device}', value)

    for device in df.index:
        writing.mean(f'gpu/memory/{device}', 100*df.loc[device, 'memused']/df.loc[device, 'memtotal'])


import av
from io import BytesIO
import numpy as np
import base64
from IPython.display import display, HTML
from pathlib import Path
from multiprocessing import cpu_count
import matplotlib.pyplot as plt
from tqdm.auto import tqdm
from .parallel import parallel
import logging
import time
import multiprocessing
from matplotlib import tight_bbox
import numbers
import sys

log = logging.getLogger(__name__)

def adjust_bbox(fig):
    bbox = fig.get_tightbbox(fig.canvas.get_renderer())
    tight_bbox.adjust_bbox(fig, bbox, fig.canvas.fixed_dpi)

def array(fig):
    adjust_bbox(fig)
    fig.canvas.draw_idle()
    renderer = fig.canvas.get_renderer()
    w, h = int(renderer.width), int(renderer.height)
    # Resolution must even, else libx264 gets upset
    h2, w2 = 2*(h//2), 2*(w//2)
    return (np.frombuffer(renderer.buffer_rgba(), np.uint8)
                        .reshape((h, w, 4))
                        [:h2, :w2, :3]
                        .copy())

class Encoder:

    def __init__(self, fps=20):
        """A context manager for encoding frames of video. Usually you'll want to use :class:`ParallelEncoder` instead.
        
        Typically used as ::

            with Encoder() as encoder:
                # Call it with each frame in turn.
                for frame in frames:
                    encoder(frame)

            # Now write it out.
            with open('test.mp4', 'b') as f:
                f.write(encoder.value)

        In this example, ``frame`` is a (H, W, 1 or 3)-dim numpy array, or a matplotlib figure.
        
        This follows the `PyAV cookbook <http://docs.mikeboers.com/pyav/develop/cookbook/numpy.html#generating-video>`_.
        """
        self._fps = fps
        self._initialized = False

    def _initialize(self, arr):
        self._content = BytesIO()
        self._container = av.open(self._content, 'w', 'mp4')

        self._stream = self._container.add_stream('h264', rate=self._fps)
        self._stream.pix_fmt = 'yuv420p'
        self._stream.height = arr.shape[0]
        self._stream.width = arr.shape[1]

        pixelformats = {1: 'gray', 3: 'rgb24'}
        self._pixelformat = pixelformats[arr.shape[2]]

        self.height, self.width = arr.shape[:2]
        self.mimetype = 'mp4'

        self._initialized = True
    
    def __enter__(self):
        return self

    def __call__(self, arr):
        if isinstance(arr, plt.Figure):
            fig = arr
            arr = array(fig)
            fig.gcf()

        if not self._initialized:
            self._initialize(arr)

        # Float arrs are assumed to have a domain of [0, 1], for backward-compatability with OpenCV.
        if np.issubdtype(arr.dtype, np.floating):
            arr = (255*arr)
        if not np.issubdtype(arr.dtype, np.uint8):
            arr = arr.astype(np.uint8).clip(0, 255)

        frame = av.VideoFrame.from_ndarray(arr, format=self._pixelformat)
        self._container.mux(self._stream.encode(frame))

    def __exit__(self, type, value, traceback):
        # Flushing the stream here causes a deprecation warning in ffmpeg
        # https://ffmpeg.zeranoe.com/forum/viewtopic.php?t=3678
        # It's old and benign and possibly only apparent in homebrew-installed ffmpeg?
        if not type:
            if hasattr(self, '_container'):
                self._container.mux(self._stream.encode())
                self._container.close()
                self.value = self._content.getvalue()
        
def html_tag(video, height=None, **kwargs):
    video = video.value if isinstance(video, Encoder) else video
    style = f'style="height: {height}px"' if height else ''
    b64 = base64.b64encode(video).decode('utf-8')
    return f"""
<video controls autoplay loop {style}>
    <source type="video/mp4" src="data:video/mp4;base64,{b64}">
    Your browser does not support the video tag.
</video>"""

def notebook(video, height=640):
    return display(HTML(html_tag(video, height)))

def _init():
    # Suppress keyboard interrupt of workers, since exiting the context 
    # manager in the parent will shut them down.
    import signal
    signal.signal(signal.SIGINT, lambda h, f: None)

def _array(f, *args, **kwargs):
    result = f(*args, **kwargs)
    if isinstance(result, plt.Figure):
        arr = array(result)
        plt.close(result)
        return arr
    else:
        return result

class ParallelEncoder:

    def __init__(self, f, fps=20, N=None):
        """A context manager for encoding frames of video in parallel. Typically used as ::
        
            with ParallelEncoder(f) as encoder:
                for x in xs:
                    encoder(x)
            encoder.notebook()  # to display the video in your notebook
            encoder.save(path)  # to save the video
        
        In this example, ``f`` is a function that takes some arguments and returns a (H, W, 1 or 3)-dim numpy array, 
        or a matplotlib figure. Whatever you call ``encoder`` with will be forwarded to ``f`` in a separate process,
        and the resulting array will be brought back to this process for encoding.

        This aligns with the common scenario where generating each frame with matplotlib is much slower than actually
        getting the arguments needed to do the generation, or doing the encoding itself. 

        :param fps: The framerate. Defaults to 20.
        :type fps: int
        :param N: The number of processes to use. Can be an integer or a float indicating the fraction of CPUs to
            use. Defaults to using 1/2 the CPUs.
        :type N: int, float
        """
        cpus = multiprocessing.cpu_count()
        if N is None:
            N = cpus//2
        elif isinstance(N, numbers.Integral):
            N = N
        elif isinstance(N, numbers.Real):
            N = int(cpus*N)
        else:
            raise ValueError(f'Number of processes must be an integer, a float, or None. Got a "{type(N)}"')

        self._encoder = Encoder(fps)
        self._f = f
        self._queuelen = N

        # Only Python >=3.7 has initializers for the process pool. 
        if sys.version_info[1] < 7:
            self._pool = parallel(_array, progress=False, N=N)
        else:
            self._pool = parallel(_array, progress=False, N=N, initializer=_init)

    def __enter__(self):
        self._futures = {}
        self._submitted = 0
        self._contiguous = 0

        self._encoder.__enter__()
        self._submit = self._pool.__enter__()
        return self

    def _process_done(self):
        while True:
            if (self._contiguous in self._futures) and self._futures[self._contiguous].done():
                result = self._futures[self._contiguous].result()
                self._encoder(result)
                del self._futures[self._contiguous]
                self._contiguous += 1
            else:
                break

    def _wait(self):
        while self._futures:
            self._process_done()
            time.sleep(.01)

    def __exit__(self, t, v, tb):
        self._wait()
        self._encoder.__exit__(t, v, tb)
        self._pool.__exit__(t, v, tb)

    def __call__(self, *args, **kwargs):
        while len(self._futures) > self._queuelen:
            self._process_done()

        self._futures[self._submitted] = self._submit(self._f, *args, **kwargs)
        self._submitted += 1
        self._process_done()

    def result(self):
        self._wait()
        return self._encoder.value

    def notebook(self):
        return notebook(self.result())

    def save(self, path):
        Path(path).write_bytes(self.result())